from hoplite.utils import server_logging
from hoplite.api import create_app
from hoplite.client.remote_job_manager import RemoteJobManager
from hoplite.exceptions import JobFailedError
import traceback
from hoplite.serializer import hoplite_loads
import argparse
import pprint
import sys
from datetime import timedelta
from tornado.wsgi import WSGIContainer
from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from .globals import HopliteServerSettings

logger = server_logging.get_server_logger(__name__)

pp = pprint.PrettyPrinter(indent=4)


def job_status(manager, args):
    job = manager.get_job(args.uuid)
    print(("Config: {0}".format(job.config())))
    try:
        print(("Status: {0}".format(job.status())))
    except JobFailedError as e:
        print("Job raised an exception during execution.")
        print(("Exception Type: {0}".format(e.type_string)))
        print(("Exception Message: {0}".format(e.msg)))
        print("Traceback:")
        for line in traceback.format_list(e.traceback):
            print(line)
    print(("Running: {0}".format(job.running())))
    print(("Finished: {0}".format(job.finished())))


def start_job(manager, args):
    job = manager.get_job(args.uuid)
    job.start()
    print("Job Started")


def create_job(manager, args):
    try:
        file = args.job_config
        json_data = open(file).read()
        config = hoplite_loads(json_data)
    except ValueError as e:
        print(("JSON parsing error: {0}".format(e)))
        return
    except Exception as e:
        config = args.job_config
        print("Could not read file, assuming you passed in config as a string.")

    job = manager.create_job(args.job_name, config, port=args.port)
    if args.start:
        job.start()
    print(("UUID: {0}".format(job.uuid)))


def list_job_plugins(manager, args):
    print("Current Job Plugins:")
    pp.pprint(manager.get_job_plugins())


def get_running_jobs(manager, args):
    print("Currently running jobs:")
    pp.pprint(manager.get_running_jobs())


def reload_plugins(manager, args):
    plugins = manager.reload_job_plugins()
    print("Current Job Plugins:")
    pp.pprint(plugins)


def get_client_options_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument('-a', '--address', default="localhost", help='The address of the hoplite server'
                                                                          ' you want to target')
    parser.add_argument('-p', '--port', default=5000, help='The port the hoplite server is listening on')

    subparsers = parser.add_subparsers(help='sub-command help')
    list_parser = subparsers.add_parser('list', help='list all available job builtin_plugins')
    list_parser.set_defaults(func=list_job_plugins)

    create_parser = subparsers.add_parser('create', help='Create a job'
                                                         '(config can be a string, or a path to a json file)')
    create_parser.add_argument('-s', '--start', action='store_true', help='starts the job after it is created')
    create_parser.add_argument('job_name', help='The namespaced name of the job you want to create')
    create_parser.add_argument('job_config', help='Path to a json file you would like '
                                                  'passed into the job for its configuration')
    create_parser.set_defaults(func=create_job)

    start_parser = subparsers.add_parser('start', help='Start a job')
    start_parser.add_argument('uuid', help='uuid of the job')
    start_parser.set_defaults(func=start_job)

    status_parser = subparsers.add_parser('info', help='Get information about a job')
    status_parser.add_argument('uuid', help='uuid of the job')
    status_parser.set_defaults(func=job_status)

    running_parser = subparsers.add_parser('running', help='Get a list of all running jobs')
    running_parser.set_defaults(func=get_running_jobs)

    reload_parser = subparsers.add_parser('reload', help='Search all the paths in sys.path and '
                                                         'reload all available plugins registered under '
                                                         'the hoplite.jobs entry point')
    reload_parser.set_defaults(func=reload_plugins)

    return parser


def get_server_options_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument('-p', '--port', default='5000', help='The port number to listen on')
    parser.add_argument('-d', '--debug', action='store_true', help='Start the server in debug mode')

    return parser


def server_main(args=sys.argv):
    parser = get_server_options_parser()
    # sys.argv includes the path of invocation as the first index in the list
    # argparse expects just the parameters if we pass a list into parse_args
    # so here we get rid of that path parameter and are left with only the args
    # from the command line
    if args == sys.argv:
        args = args[1:]
    args = parser.parse_args(args)

    HopliteServerSettings.debug = args.debug

    app = create_app()
    logger.info('Starting Hoplite server on port {}'.format(args.port))
    http_server = HTTPServer(WSGIContainer(app))
    http_server.listen(args.port)
    ioloop = IOLoop.instance()
    # This is needed to ensure Ctrl-C kills the server quickly
    set_ping(ioloop, timedelta(seconds=2))
    try:
        ioloop.start()
    except:
        raise
    finally:
        # Ensure that the server always closes the socket when it's no longer
        # in use
        ioloop.stop()


def client_main():
    parser = get_client_options_parser()
    args = parser.parse_args()
    manager = RemoteJobManager(args.address, args.port)
    args.func(manager, args)


# I think this periodically pings the tornado server. The reason this is needed
# is because Tornado apparently doesn't process the Ctrl-C exception until the
# next request after the Ctrl-C is received.
def set_ping(ioloop, timeout):
    ioloop.add_timeout(timeout, lambda: set_ping(ioloop, timeout))



import logging
import os
import random
import tempfile
import time

from exceptions import ExternalCustomError, ExternalEmptyError
from hoplite.remote_enabler import remotify

logger = logging.getLogger(__name__)


@remotify(__name__)
class TestClass:
    class MyPrivateError(Exception):
        pass

    def __init__(self, string_value=None, number_value=None):
        self.string_value = string_value
        self.number_value = number_value
        self.file_contents = "".join([random.choice('abcdefghijklmnopqrstuvwxyz') for _ in xrange(8)])

    def return_none(self):
        """
        Used to test that a None value can be returned properly
        :return: None
        """
        None

    def return_number(self):
        """
        Used to test that a single value can be returned properly
        :return: A number, 777
        """
        return 777

    def return_tuple(self):
        """
        Used to test that a tuple can be returned properly
        :return: A tuple containing a string, an int, and a float
        """
        return 'All', 4, 1.11

    def return_single_list(self):
        """
        Used to test that a single list can be returned properly
        :return: A list containing several strings
        """
        return ['This', 'is', 'a', 'list']

    def do_math(self, number_1, number_2):
        """
        Used to test that numeric arguments can be passed through as arguments
        :param number_1: Arbitrary number
        :param number_2: Arbitrary number
        :return: The product of the two arguments
        """
        return number_1 * number_2

    def pass_common_class(self, date):
        """
        Used to test that more complex class objects can be passed through as arguments
        :param date: Datetime object
        :return: A tuple containing the information stored in the datetime argument
        """
        return date.year, date.month, date.day, date.hour, date.minute, date.second, date.microsecond

    def pass_custom_class(self, dummy):
        """
        Used to test that custom class objects can be passed through as arguments
        :param dummy: DummyClass
        :return: The sum of numbers contained in the class
        """
        return dummy.add_numbers()

    def return_multiple_lists(self):
        """
        Used to test that a jagged, 2-dimensional list can be returned properly
        :return: A list of lists, each of which are different lengths and contain different values
        """
        return ['This', 'is', 1, 'list'], ['and', 'this', 'is', 'another', 'list']

    def get_class_variables(self):
        """
        Used to test that class member variables are transmitted to the remote
        function and then can be manipulated and returned properly
        :return: A string, and a number multiplied by two, both class variables
        """
        return self.string_value, self.number_value * 2

    def raise_type_error(self):
        """
        Test that normal exceptions can be raised properly
        :raise TypeError: Unconditionally raises Type Error
        """
        raise TypeError('___Failure1')

    def raise_custom_error(self):
        """
        Used to test that custom exceptions can be raised properly
        :raise ExternalEmptyError: Unconditionally raises Weird Error
        """
        raise ExternalEmptyError('___Failure2')

    def create_file(self):
        """
        Used to test remote file creation using a string stored in class instance
        :return: The name of the random temp file that was created
        """
        handle, filename = tempfile.mkstemp()
        with os.fdopen(handle, 'w') as fout:
            fout.write(self.file_contents)
        return filename

    def create_specified_file(self, filename, contents='default_contents'):
        """
        Used to test remote file creation using a string passed into the function. The filename
        is also passed in. This is specifically used to test the asynchronous functionality
        of the metaclass.
        """
        time.sleep(1)
        with open(filename, 'w') as fout:
            fout.write(contents)

    def call_nested_function(self):
        """
        Used to test that exception handling works properly when jobs are nested. This function
        starts a job that raises a type error.
        :raise TypeError: Raised when the function remote_raise_type_error is called
        """
        self.remote_raise_type_error('localhost:5001')

    def raise_private_error(self):
        """
        Used to test that private errors defined within a class cannot be properly pickled. This means
        that the exception should come out on the client side as a JobFailedError instead.
        :raise MyPrivateException: A private exception class
        """
        raise TestClass.MyPrivateError('___Failure3')

    def raise_public_error(self):
        """
        Used to test that externally defined, custom errors can be raised on the client side.
        :raise ExternalCustomError: Unconditionally raises ExternalCustomError
        """
        raise ExternalCustomError('___Failure4')

    def return_custom_exception(self):
        """
        Used to test that custom class instances (such as exceptions) can be returned
        :return: Returns an ExternalCustomError exception object
        """
        return ExternalCustomError('___Failure4')

    def long_job(self):
        """
        Used to test job timeouts on remotely called functions
        """
        time.sleep(7)

    def log_normal(self, dummy_var_1, dummy_var_2=None):
        logger.info('Logging in log_normal function')
        return str(dummy_var_1) + ' + ' + str(dummy_var_2)

    def log_nested_caller(self):
        logger.info('Currently in class caller function')
        self.log_nested_callee()

    def log_nested_callee(self):
        logger.info('Currently in class callee function')

    @staticmethod
    def static_return_number():
        """
        Used to test that static methods can be called remotely
        :return: A number, 999
        """
        return 999


class ParentClass(object):
    """
    Normal new-style class which doesn't do much at all.
    """
    def __init__(self, parent_value):
        self.parent_value = parent_value

    def func_1(self, value_to_add):
        return 1 + value_to_add + self.parent_value

    def func_2(self, value_to_add):
        return 2 + value_to_add + self.parent_value


@remotify(__name__)
class ChildClass(ParentClass):
    """
    Class which inherits from the ParentClass class. This is to make sure
    that it is possible to have a parent which doesn't use the metaclass, provided
    it inherits from object. In addition, this class exercises various features which
    are used by the metaclass, such as function arguments and return values. Also, values
    from ParentClass are used in calculating the function return values, in order to test
    that inherited classes don't get broken by the remotable system.
    """
    def __init__(self, child_value, parent_value):
        self.child_value = child_value
        ParentClass.__init__(self, parent_value)

    def func_3(self, value_to_add):
        return 3 + value_to_add + self.child_value + self.parent_value

    def func_4(self, value_to_add):
        return 4 + value_to_add + self.child_value + self.parent_value


class GrandchildClass(ChildClass):
    """
    Class which inherits from the ChildClass class. This is to make sure that the metaclass
    applied to ChildClass also gets applied to this class, GrandchildClass. Also, values
    from ChildClass and ParentClass are used in calculating the function return values, in
    order to test that inherited classes don't get broken by the remotable system.
    """
    def __init__(self, grandchild_value, child_value, parent_value):
        self.grandchild_value = grandchild_value
        ChildClass.__init__(self, child_value, parent_value)

    def func_5(self, value_to_add):
        return 5 + value_to_add + self.grandchild_value + self.child_value + self.parent_value

    def func_6(self, value_to_add):
        return 6 + value_to_add + self.grandchild_value + self.child_value + self.parent_value


import socket
import time
from hoplite.exceptions import TimeoutError

from urllib.parse import urlparse

from io import BytesIO
import pip
import sys
from subprocess import Popen, PIPE

from hoplite.remote_enabler import remotify

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def is_hoplite_available(address, port=5000):
    """
    Find if the hoplite process is accessible on the remote target

    :return: True if hoplite is available and listening on the target
    :rtype: Boolean
    """
    if ':' in address:
        temp_address = address.split(':')[0]
        port = address.split(':')[1]
        address = temp_address

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((address, port))
        s.close()
    except socket.error:
        return False
    return True


def wait_for_hoplite(address, port=5000, retry_period_s=10, retries=30):
    """
    Waits for Hoplite to be active. If it finds hoplite is active before
    the timeout.

    :param int retry_period_s: The number of seconds between retying to connect
        to Hoplite
    :param int retries: The number of attempts to connect to Hoplite to make
        before throwing an exception
    """
    estimated_units = "seconds"
    estimated_timeout = retry_period_s * retries
    if estimated_timeout >= 60:
        estimated_timeout /= 60
        estimated_units = "minutes"

    if ':' in address:
        temp_address = address.split(':')[0]
        port = address.split(':')[1]
        address = temp_address

    logger.info("Waiting for hoplite to become available on {0}:{1}".format(address, port))

    while True:
        if is_hoplite_available(address, port):
            logger.info("Hoplite is online at {0}:{1}".format(address, port))
            break
        time.sleep(retry_period_s)
        retries -= 1
        if retries <= 0:
            errmsg = 'Hoplite Server could not be reached on host "{0}:{1}" after waiting for at least {2} {3}'.format(
                     address,
                     port,
                     estimated_timeout,
                     estimated_units)
            logger.error(errmsg)
            raise TimeoutError(errmsg)


@remotify(__name__)
def install_python_package(package, server=None, options=None):
    """
    Installs a python package via pip
    :param package: Name of python package
    :param server: Server for pip to retrieve from. If no server is specified,
        then the default server will be used based on the pip configuration
        file in the user folder. Note that any server specified will be assumed
        to be a trusted host
    :param options: A list of options to use when installing the package.
    :type options: list of [str]
    :return: The output from running pip
    """
    from pip.status_codes import SUCCESS, ERROR, UNKNOWN_ERROR, VIRTUALENV_NOT_FOUND, PREVIOUS_BUILD_DIR_ERROR
    pip_error_codes = {
        ERROR: 'ERROR',
        UNKNOWN_ERROR: 'UNKNOWN_ERROR',
        VIRTUALENV_NOT_FOUND: 'VIRTUALENV_NOT_FOUND',
        PREVIOUS_BUILD_DIR_ERROR: 'PREVIOUS_BUILD_DIR_ERROR'
    }

    args = ['install', package]

    if options is not None:
        args += options

    if server is not None:
        logger.info('Installing {} from server: {}'.format(package, server))
        trusted_host = urlparse(server)[1]
        if ':' in trusted_host:
            trusted_host = trusted_host.split(':')[0]
        args += ['-i', server, '--trusted-host', trusted_host]
    else:
        logger.info('Installing {} from default Pypi server'.format(package))

    stdout_backup = sys.stdout
    stderr_backup = sys.stderr
    with BytesIO() as bytes_io:
        sys.stdout = bytes_io
        sys.stderr = bytes_io
        exit_code = pip.main(args)
        pip_output = bytes_io.getvalue()
    sys.stdout = stdout_backup
    sys.stderr = stderr_backup
    if exit_code == SUCCESS:
        return pip_output
    elif pip_error_codes.get(exit_code) is not None:
        # We could raise PipError, but it seems like it's intended only for
        # internal pip use.
        raise RuntimeError(
            'Pip exited with nonzero exit code: {} and output: {}'.format(
                pip_error_codes[exit_code], pip_output))
    else:
        raise RuntimeError(
            'Pip encountered an error while installing. Output: {}'.format(
                pip_output))


@remotify(__name__)
def uninstall_python_package(package):
    logger.info("Uninstalling %s with pip" % package)
    pip.main(['uninstall', package, '-y'])


@remotify(__name__)
def run_cmd_command(*args, **kwargs):
    """
    Runs a subprocess command.
    """
    if "stdout" in kwargs:
        kwargs.pop("stdout")
    p = Popen(*args, stdout=PIPE, stderr=PIPE, **kwargs)
    std_out, std_err = p.communicate()
    err_code = p.returncode
    return std_out, std_err, err_code



import shutil
import os
import platform
from subprocess import Popen, PIPE, STDOUT


def authenticate_network_resource(resource, username, password):
    """ Authenticate a network share on the local machine
    @param resource: The UNC path for the network share
    @param username: The username to use. Domain names should be prefixed like
        this r'AMER\nitest'
    @param password: The password for user
    @return: None
    """

    opsys = platform.system().lower()

    if opsys == 'windows':
        _authenticate_network_resource_windows(
            resource=resource, username=username, password=password)
    else:
        raise NotImplementedError(
            'Support for this method on %s is not yet implemented.' % opsys)


def _authenticate_network_resource_windows(resource, username, password):
    unc, path = os.path.splitunc(resource)
    if username is not None and password is not None:
        cmdList = ["NET", "USE", unc, "/USER:{0}".format(username), password]
    else:
        cmdList = ["NET", "USE", unc]

    proc = Popen(cmdList, stdout=PIPE, stderr=STDOUT, shell=True)
    rtn = proc.communicate()[0]
    if 0 != proc.returncode:
        error = RuntimeError(
            "Error %d from %s'\n%s" % (
                proc.returncode, " ".join(cmdList), rtn))
        error.returncode = proc.returncode
        raise error
    return rtn


def download_network_folder(network_path, local_path, user, password):
    """
    Downloads the directory at network_path into local_path
    The folder in local_path must not exist
    """
    authenticate_network_resource(network_path, user, password)
    shutil.copytree(network_path, local_path)


def upload_folder_to_network_directory(network_path, local_path, user, password):
    """
    Uploads the folder at local_path to the network_path
    The upload location in network_path must not exist
    """
    authenticate_network_resource(network_path, user, password)
    shutil.copytree(local_path, network_path)



"""
Role
====
Used to manage jobs on a remote hoplite server. This can be used to issue
commands to the job or get info about the job.
This is a wrapper around most of the commands in :ref:`REST-API-Jobs`

API
===
"""
import pickle
import time

from hoplite.client.helpers import ClientMixin
from hoplite.exceptions import (
    JobDoesNotExistError,
    TimeoutError,
    ConnectionError,
    JobFailedError)
from hoplite.serializer import hoplite_loads
import requests.exceptions


class RemoteJob(ClientMixin):
    """
    The representation of a job on a remote hoplite server
    """

    def __init__(self, address, port=5000, name="", uuid="", api_key="", config={}):
        """
        :param address: IP address or hostname of the computer running the job.
            If desired, the address may be in the form "address:port", rather
            than specifying the port in the second parameter
        :param port: Port of the remote computer hoplite is listening on. This
            is ignored if the address includes the port
        :param name: qualified name of the job to run
        :param uuid: the uuid of the job. If left blank then job will be
            created by the daemon running on addr
        :type uuid: string or None
        :param api_key: this is used to only allow the running job to update
            its own status
        :raises: InvalidAddressError
        :raises: JobDoesNotExistError
        :raises: ConnectionError
        """
        if ':' in address:
            self.address = address.split(':')[0]
            self.port = address.split(':')[1]
        else:
            self.address = address
            self.port = port
        self._daemon_addr = 'http://{0}:{1}'.format(self.address, self.port)
        self._config = config
        self.name = name
        self.uuid = uuid
        self._api_key = api_key
        self._last_poll = 0

        try:
            if not self.uuid:
                self._create_job()
            self._get_job()
        except requests.exceptions.ConnectionError:
            raise ConnectionError(self.address)

    def config(self, force=False):
        """
        Get the config dictionary for this job

        :return: the configuration dictionary the job was created with
        :rtype: dict
        """
        self._get_job(force)
        return self._config

    def status(self, force=False):
        """
        Get the status dictionary of the job

        :return: the status dictionary of the job
        :rtype: dict
        :raises JobFailedError: if the job raised an exception
        """
        self._get_job(force)
        exception_dict = self._status.get("exception", None)
        if exception_dict:
            # Raise an exception, being sure to store the information from the
            # previous exception. This is so that stack traces for every job
            # are preserved, in the event that jobs call other jobs which then
            # throw exceptions.
            raise JobFailedError(
                self.address,
                self.uuid,
                pickle.loads(exception_dict['traceback']),
                exception_dict['previous_exception'])
        return self._status

    def start(self):
        """
        Start the job

        :return: true if the job was started successfully
        :raises JobDoesNotExistError: Job does not exist on the server
        """
        self._get_job()
        resp = self.jput(
            self._daemon_addr + '/jobs/{0}/start'.format(self.uuid))
        return resp.json()["started"]

    def join(self, timeout=-1):
        """
        This will block until the job is finished and then return True.
        Timeout is in seconds.

        Blocks infinitely by default (timeout=-1).

        :raises TimeoutError: The specified timeout was reached
        :raises JobFailedError: The job threw an exception
        """
        num_seconds = 0
        poll_interval = .05
        while num_seconds < timeout or timeout == -1:
            if self.finished():
                return True
            time.sleep(poll_interval)
            num_seconds += poll_interval
        raise TimeoutError(self.uuid)

    def kill(self, force=False):
        """
        Kill the job

        :return: true if the kill command was sent successfully. Success here
            does not mean the job is stopped, it only means that a kill signal
            was successfully sent to the job
        :rtype: bool
        :raises JobDoesNotExistError: if the job does not exist on the targeted
            server
        """
        self._get_job(force)
        resp = self.jput(
            self._daemon_addr + '/jobs/{0}/kill'.format(self.uuid))
        return hoplite_loads(str(resp.text))["killed"]

    def running(self, force=False):
        """
        :return: true if the job is currently executing on the target machine
        :rtype: bool
        :raises JobDoesNotExistError: Job not found on the server
        """
        self._get_job(force)
        return self._running

    def finished(self, force=False):
        """
        Returns true if the job on the target machine is no longer executing

        :return: If the job has run and is currently not running
        :rtype: bool
        :raises JobDoesNotExistError: Job not found on the server
        :raises JobFailedError: Job raised an exception
        """
        self.status(force)
        return self._finished

    def _get_job(self, force=False):
        """

        I call this before most other requests to get the status code sanity
        check.
        This method is rate limited for sanity
        """
        time_elapsed = time.time() - self._last_poll
        if time_elapsed > .2 or force:
            resp = self.jget(self._daemon_addr + '/jobs/{0}'.format(self.uuid))
            if resp.status_code == 404:
                raise JobDoesNotExistError
            self._set_attributes_from_response_json(
                hoplite_loads(str(resp.text)))
            self._last_poll = time.time()

    def _create_job(self):
        job_data = {"name": self.name, "config": self._config, "port": self.port}
        resp = self.jpost(self._daemon_addr + '/jobs', data=job_data)
        if resp.status_code == 400:
            raise JobDoesNotExistError(hoplite_loads(str(resp.text))["error"])
        self._set_attributes_from_response_json(hoplite_loads(str(resp.text)))

    def _set_attributes_from_response_json(self, resp_dict):
        job = resp_dict
        self.uuid = job["uuid"]
        self.name = job["name"]
        self._status = job.get("status", {})
        self._config = job.get("config", {})
        self._running = job.get("running", False)
        self._finished = job.get("finished", False)



import errno
import logging
import logging.config
import logging.handlers
import os
import platform
import string

# Tone down logging from the http modules
log = logging.getLogger('werkzeug')
log.setLevel(logging.WARNING)
log = logging.getLogger('requests')
log.setLevel(logging.WARNING)
log = logging.getLogger('tornado')
log.setLevel(logging.WARNING)


def get_server_logger(name=""):
    # set up hoplite_logging
    logger = _get_base_logger(name)
    return logger


def get_job_logger(name="", uuid=""):
    logger = _get_base_logger("hoplite.{0}_{1}".format(name, uuid))

    log_file = "{0}_{1}.log".format(name, uuid)
    log_file_path = log_file

    opsys = platform.system().lower()
    if opsys == 'windows':
        log_file_path = os.path.join(r"C:\logs\hoplite\jobs", name)
        try:
            os.makedirs(log_file_path)
        except OSError as e:
            if not (e.errno == errno.EEXIST and os.path.isdir(log_file_path)):
                raise
        log_file_path = os.path.join(log_file_path, log_file)
    elif opsys == 'linux' or opsys == 'darwin':
        log_file_path = os.path.join(
            os.path.expanduser('~'), '.hoplite', 'jobs', name)
        try:
            os.makedirs(log_file_path)
        except OSError as e:
            if not (e.errno == errno.EEXIST and os.path.isdir(log_file_path)):
                raise
        log_file_path = os.path.join(log_file_path, log_file)

    fh = logging.handlers.RotatingFileHandler(
        filename=log_file_path,
        maxBytes=2097152,
        backupCount=100)
    fh.setLevel(logging.DEBUG)
    # Add context info
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    return logger


def _get_base_logger(name=""):
    opsys = platform.system().lower()
    if opsys == 'windows':
        log_file_path = os.path.join(r"C:\logs\hoplite\hoplite")
    elif opsys == 'linux' or opsys == 'darwin':
        log_file_path = os.path.join(
            os.path.expanduser('~'), '.hoplite', 'hoplite')
    else:
        raise NotImplementedError()

    # Create the directory if it doesn't exist
    try:
        os.makedirs(log_file_path)
    except OSError as e:
        if not (e.errno == errno.EEXIST and os.path.isdir(log_file_path)):
            raise

    logging.config.fileConfig(
        os.path.join(
            os.path.split(__file__)[0], "default{0}.cfg".format(opsys)),
        disable_existing_loggers=False
    )
    logger = logging.getLogger(name)
    return logger


def add_remote_function_logging_handlers(function_namespace, timestamp, filename=None):
    logger = logging.getLogger()

    # If this function has already been called in this process
    if len(logger.handlers) >= 2:
        return

    # This function gets called twice for each remote function call. On the
    # first call, filename is not provided. On the second call, the filename
    # from last time is passed in so the same log filename is used.
    if filename is None:
        full_log_name = get_log_filename(function_namespace, timestamp)
    else:
        full_log_name = filename

    logger.setLevel(logging.DEBUG)

    hoplite_remote_enabler_fh = logging.FileHandler(full_log_name)
    hoplite_remote_enabler_fh.setLevel(logging.DEBUG)
    hoplite_remote_enabler_fh.setFormatter(get_formatter_verbose())

    hoplite_remote_enabler_sh = logging.StreamHandler()
    hoplite_remote_enabler_sh.setLevel(logging.INFO)
    hoplite_remote_enabler_sh.setFormatter(get_formatter_limited())

    logger.addHandler(hoplite_remote_enabler_fh)
    logger.addHandler(hoplite_remote_enabler_sh)

    return full_log_name


def get_log_filename(function_namespace, timestamp):
    namespace_split = string.split(function_namespace, '.')
    function_name = namespace_split.pop()
    folder_hierarchy = os.path.join(*namespace_split)
    opsys = platform.system().lower()
    if opsys == 'windows':
        log_root_path = os.path.join(r'C:\logs\hoplite\remoted_functions')
    elif opsys == 'linux' or opsys == 'darwin':
        log_root_path = os.path.join(
            os.path.expanduser('~'), '.hoplite', 'remoted_functions')
    else:
        raise NotImplementedError()
    log_root_path = os.path.join(log_root_path, folder_hierarchy)
    # Create the directory if it doesn't exist
    try:
        os.makedirs(log_root_path)
    except OSError as e:
        if not (e.errno == errno.EEXIST and os.path.isdir(log_root_path)):
            raise
    # Handle the case where multiple instances of the same function are called
    # within a single second. This would otherwise result in the same filename
    # being used multiple times
    extension_num = 1
    log_file_name = os.path.join(
        log_root_path, timestamp + ' ' + function_name + ".log")
    while os.path.isfile(log_file_name):
        log_file_name = os.path.join(
            log_root_path,
            timestamp + ' ' + function_name + '_' + str(extension_num) + ".log")
        extension_num += 1

    return log_file_name


def get_formatter_verbose():
    return logging.Formatter('%(levelname)-8s '
                             '%(asctime)s '
                             '%(filename)+30s:%(lineno)-5s '
                             '%(threadName)-15s %(message)s',
                             "%I:%M:%S %p")


def get_formatter_limited():
    return logging.Formatter('%(levelname)-8s '
                             '%(asctime)s '
                             '%(filename)+20s:%(lineno)-5s '
                             '%(message).200s',
                             "%I:%M:%S %p")



import unittest2
import os
import socket
import tempfile
from pyftpdlib.authorizers import DummyAuthorizer
from pyftpdlib.handlers import FTPHandler
from pyftpdlib.servers import FTPServer
from hoplite.builtin_plugins.download_folder_from_ftp_job import run
from hoplite.builtin_plugins.constants import DownloadFolderFromFtpJobConstants as FtpKeys
from hoplite.client.status_updater import MockStatusUpdater
import threading
import shutil
import platform

# Attempt to use IP rather than hostname (test suite will run a lot faster)
try:
    HOST = socket.gethostbyname('localhost')
except socket.error:
    HOST = 'localhost'

PORT = 8976


class FTPd(threading.Thread):
    """
    A threaded FTP server used for running tests.

    This is taken from the pyftpdlib tests because we need to be able to
    run our tests on all platforms, just like they do.

    This is basically a modified version of the FTPServer class which
    wraps the polling loop into a thread. This allows tests to run
    with a FTP Server that we create and start for each test.

    The instance returned can be used to start(), stop() and
    eventually re-start() the server.
    """
    handler = FTPHandler
    server_class = FTPServer

    def __init__(self, addr=None):
        threading.Thread.__init__(self)
        self.__serving = False
        self.__stopped = False
        self.__lock = threading.Lock()
        self.__flag = threading.Event()
        if addr is None:
            addr = (HOST, PORT)

        authorizer = DummyAuthorizer()
        test_dir = os.path.split(os.path.split(__file__)[0])[0]
        ftp_dir = os.path.join(test_dir, "test_resources", "ftp_test")
        authorizer.add_anonymous(ftp_dir)
        authorizer.add_user("no_pass", "", ftp_dir)
        authorizer.add_user("with_pass", "password", ftp_dir)
        self.handler.authorizer = authorizer
        # lowering buffer sizes = more cycles to transfer data
        # = less false positive test failures
        self.handler.dtp_handler.ac_in_buffer_size = 32768
        self.handler.dtp_handler.ac_out_buffer_size = 32768
        self.server = self.server_class(addr, self.handler)
        self.host, self.port = self.server.socket.getsockname()[:2]

    def __repr__(self):
        status = [self.__class__.__module__ + "." + self.__class__.__name__]
        if self.__serving:
            status.append('active')
        else:
            status.append('inactive')
        status.append('%s:%s' % self.server.socket.getsockname()[:2])
        return '<%s at %#x>' % (' '.join(status), id(self))

    @property
    def running(self):
        return self.__serving

    def start(self, timeout=0.001):
        """Start serving until an explicit stop() request.
        Polls for shutdown every 'timeout' seconds.
        """
        if self.__serving:
            raise RuntimeError("Server already started")
        if self.__stopped:
            # ensure the server can be started again
            FTPd.__init__(self, self.server.socket.getsockname(), self.handler)
        self.__timeout = timeout
        threading.Thread.start(self)
        self.__flag.wait()

    def run(self):
        self.__serving = True
        self.__flag.set()
        while self.__serving:
            self.__lock.acquire()
            self.server.serve_forever(timeout=self.__timeout, blocking=False)
            self.__lock.release()
        self.server.close_all()

    def stop(self):
        """Stop serving (also disconnecting all currently connected
        clients) by telling the serve_forever() loop to stop and
        waits until it does.
        """
        if not self.__serving:
            raise RuntimeError("Server not started yet")
        self.__serving = False
        self.__stopped = True
        self.join()


class TestDownloadFolderFromFTPJob(unittest2.TestCase):
    server_class = FTPd

    def setUp(self):
        self.server = self.server_class()
        self.server.start()

        test_dir = os.path.split(os.path.split(__file__)[0])[0]
        self.temp_dir = tempfile.mkdtemp(dir=os.path.join(test_dir, "test_resources"))
        self.config = {
            FtpKeys.SERVER_ADDRESS: HOST,
            FtpKeys.SERVER_PORT: PORT,
            FtpKeys.FTP_ROOT: "/",
            FtpKeys.DEST_ROOT: self.temp_dir
        }

    def tearDown(self):
        self.server.stop()
        self.server.join()
        shutil.rmtree(self.temp_dir)

    def test_copies_folder_anonymous(self):
        self._run_job(self.config)
        self._validate_directory(self.temp_dir)

    def test_copies_folder_user_no_pass(self):
        self.config[FtpKeys.USERNAME] = "no_pass"
        self._run_job(self.config)
        self._validate_directory(self.temp_dir)

    def test_copies_folder_user_pass(self):
        self.config[FtpKeys.USERNAME] = "with_pass"
        self.config[FtpKeys.PASSWORD] = "password"
        self._run_job(self.config)
        self._validate_directory(self.temp_dir)

    def test_errors_if_remote_dir_doesnt_exist(self):
        self.config[FtpKeys.FTP_ROOT] = "/not/exist"
        status = self._run_job(self.config)
        self.assertEqual(status.status, {"error": "Could not open the ftp directory: '/not/exist'"})

    def test_errors_if_local_dir_doesnt_exist(self):
        self.config[FtpKeys.DEST_ROOT] = "/fake/path"
        status = self._run_job(self.config)
        self.assertEqual(status.status, {"error": "Local path does not exist: '/fake/path'"})

    def test_remote_server_does_not_exist(self):
        self.config[FtpKeys.SERVER_ADDRESS] = "not-real-address"
        status = self._run_job(self.config)

        opsys = platform.system().lower()
        if opsys == 'windows':
            self.assertEqual(status.status, {"error": "[Errno 11004] getaddrinfo failed"})
        elif opsys == 'linux':
            self.assertEqual(status.status, {'error': '[Errno -2] Name or service not known'})
        elif opsys == 'darwin':
            self.assertEqual(status.status, {'error': '[Errno 8] nodename nor servname provided, or not known'})

    def test_file_already_exists_on_local_overwrites(self):
        # stat()[-2] is the time of mose recent content modification
        self._run_job(self.config)
        first_mod_date = os.stat(os.path.join(self.temp_dir, "test_file_1.txt"))[-2]
        self._run_job(self.config)
        second_mod_date = os.stat(os.path.join(self.temp_dir, "test_file_1.txt"))[-2]
        self.assertGreaterEqual(second_mod_date, first_mod_date)

    def _validate_directory(self, directory):
        file_1_path = os.path.join(directory, "test_file_1.txt")
        file_2_path = os.path.join(directory, "test_directory_1", "test_file_2.txt")

        with open(file_1_path) as f:
            expected = "Hello world!"
            actual = f.read()
            self.assertEqual(expected, actual)

        with open(file_2_path) as f:
            expected = "Hello moto!"
            actual = f.read()
            self.assertEqual(expected, actual)

    def _run_job(self, config):
        status = MockStatusUpdater()
        run(config, status)
        return status


from hoplite.serializer import hoplite_dumps


class StatusCodeTestMixin(object):
    def assertStatusCode(self, response, status_code):
        """Assert the status code of a Flask test client response

        :param response: The test client response object
        :param status_code: The expected status code
        """
        self.assertEquals(status_code, response.status_code)
        return response

    def assertOk(self, response):
        """Test that response status code is 200

        :param response: The test client response object
        """
        return self.assertStatusCode(response, 200)

    def assertBadRequest(self, response):
        """Test that response status code is 400

        :param response: The test client response object
        """
        return self.assertStatusCode(response, 400)

    def assertForbidden(self, response):
        """Test that response status code is 403

        :param response: The test client response object
        """
        return self.assertStatusCode(response, 403)

    def assertNotFound(self, response):
        """Test that response status code is 404

        :param response: The test client response object
        """
        return self.assertStatusCode(response, 404)

    def assertContentType(self, response, content_type):
        """Assert the content-type of a Flask test client response

        :param response: The test client response object
        :param content_type: The expected content type
        """
        self.assertEquals(content_type, response.headers['Content-Type'])
        return response

    def assertOkHtml(self, response):
        """Assert the response status code is 200 and an HTML response

        :param response: The test client response object
        """
        return self.assertOk(
            self.assertContentType(response, 'text/html; charset=utf-8'))

    def assertJson(self, response):
        """Test that content returned is in JSON format

        :param response: The test client response object
        """
        return self.assertContentType(response, 'application/json')

    def assertOkJson(self, response):
        """Assert the response status code is 200 and a JSON response

        :param response: The test client response object
        """
        return self.assertOk(self.assertJson(response))

    def assertBadJson(self, response):
        """Assert the response status code is 400 and a JSON response

        :param response: The test client response object
        """
        return self.assertBadRequest(self.assertJson(response))


class FlaskTestCaseMixin(object):
    def _json_data(self, kwargs, csrf_enabled=True):
        if 'data' in kwargs:
            kwargs['data'] = hoplite_dumps(kwargs['data'])
        if not kwargs.get('content_type'):
            kwargs['content_type'] = 'application/json'
        return kwargs

    def _request(self, method, *args, **kwargs):
        kwargs.setdefault('content_type', 'text/html')
        kwargs.setdefault('follow_redirects', True)
        return method(*args, **kwargs)

    def _jrequest(self, *args, **kwargs):
        return self._request(*args, **kwargs)

    def get(self, *args, **kwargs):
        return self._request(self.client.get, *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self._request(self.client.delete, *args, **kwargs)

    def jget(self, *args, **kwargs):
        return self._jrequest(self.client.get, *args, **kwargs)

    def jpost(self, *args, **kwargs):
        return self._jrequest(self.client.post, *args, **self._json_data(kwargs))

    def jput(self, *args, **kwargs):
        return self._jrequest(self.client.put, *args, **self._json_data(kwargs))

    def jpatch(self, *args, **kwargs):
        return self._jrequest(self.client.patch, *args, **self._json_data(kwargs))

    def jdelete(self, *args, **kwargs):
        return self._jrequest(self.client.delete, *args, **self._json_data(kwargs))


import bson
import datetime
from httmock import urlmatch, HTTMock, response
import pickle
import re
import sys
from tblib import pickling_support
import unittest2

from hoplite.client.remote_job import RemoteJob
from hoplite.exceptions import JobFailedError, JobDoesNotExistError, TimeoutError, ConnectionError
from hoplite.utils import server_logging
from hoplite.serializer import hoplite_dumps

# This makes it so that traceback objects can be pickled
pickling_support.install()

logger = server_logging.get_server_logger(__name__)


# Get an arbitrary traceback object to use for the tests
try:
    raise TypeError()
except:
    pickled_traceback = pickle.dumps(sys.exc_info()[2])


job_dict = {
    "uuid": "correctuuid",
    "name": "test_plugins.wait_10_seconds",
    "config": {"something": "is missing"},
    "status": {
        "Roger": "Roger what's your status?",
        "anumber": 123,
        "objectid": bson.objectid.ObjectId(),
        "datetime": datetime.datetime.now(),
        "regexp": re.compile("hoplite")
    },
    "running": True,
    "finished": False
}

job_dict_2 = {
    "uuid": "correctuuid",
    "name": "test_plugins.wait_10_seconds",
    "config": {"something": "is missing"},
    "status": {"Roger": "Roger what's your status?"},
    "running": False,
    "finished": True
}

job_dict_name_something = {
    "uuid": "correctuuid",
    "name": "something",
    "config": {"something": "is missing"},
    "status": {"Roger": "Roger what's your status?"},
    "running": True,
    "finished": False
}

started_dict = {
    "uuid": "correctuuid",
    "started": "true"
}

killed_dict = {
    "uuid": "correctuuid",
    "killed": "true"
}

excepted_job = {
    "uuid": 0,
    "name": "Failing Job",
    "status": {
        "exception": {
            "traceback": pickled_traceback,
            "previous_exception": {
                "type": "TypeError",
                "message": "Wrong type",
                "exception_object": 'bla_bla_pickled_object'
            }
        }
    }
}

exception_bubble_up_job = {
    "uuid": 0,
    "name": "Failing Job",
    "status": {
        "exception": {
            "address": "10.2.1.10",
            "uuid": 5,
            "type": "TypeError",
            "message": "Wrong type",
            "traceback": pickled_traceback
        }
    }
}

@urlmatch(path='\/jobs\/\w+$')
def get_with_bubbled_up_exception(url, request):
    return response(200, hoplite_dumps(exception_bubble_up_job), {'content-type': 'application/json'})

@urlmatch(path='\/jobs\/\w+$')
def get_with_exception(url, request):
    return response(200, hoplite_dumps(excepted_job), {'content-type': 'application/json'})

@urlmatch(path='\/jobs\/\w+$')
def get_specific_job(url, request):
    return response(200, hoplite_dumps(job_dict), {'content-type': 'application/json'})

@urlmatch(netloc="localhost:5001", path='\/jobs\/\w+$')
def get_specific_job_404(url, request):
    return response(404)

@urlmatch(path='/jobs$')
def post_jobs(url, request):
    if request.method == 'POST' and request.headers['Content-type'] == 'application/json':
        return response(200, hoplite_dumps(job_dict_name_something), {'content-type': 'application/json'})

@urlmatch(path='/jobs$')
def post_jobs_400(url, request):
    if request.method == 'POST' and request.headers['Content-type'] == 'application/json':
        return response(400, hoplite_dumps({"error": "Job not found"}), {'content-type': 'application/json'})

@urlmatch(path='\/jobs\/\w+$')
def get_specific_job_named_something(url, request):
    return response(200, hoplite_dumps(job_dict_name_something), {'content-type': 'application/json'})

@urlmatch(path='\/jobs\/\w+$')
def get_specific_job_running_false_finished_true(url, request):
    return response(200,  hoplite_dumps(job_dict_2), {'content-type': 'application/json'})

@urlmatch(path='\/jobs\/\w+\/start$')
def start_job(url, request):
    if request.method == 'PUT':
        return response(200,  hoplite_dumps(started_dict), {'content-type': 'application/json'})

@urlmatch(path='\/jobs\/\w+\/kill$')
def kill_job(url, request):
    if request.method == 'PUT':
        return response(200,  hoplite_dumps(killed_dict), {'content-type': 'application/json'})


class TestRemoteJob(unittest2.TestCase):
    def setUp(self):
        with HTTMock(get_specific_job):
            self.job = RemoteJob("localhost", 5001, "test_plugins.wait_10_seconds", "correctuuid", "api_key")

    def test_init_addr_name_uuid_apikey(self):
        #TODO There is a testing hole here...Where's config?! Input and Return
        self.assertEquals(self.job.address, "localhost")
        self.assertEquals(self.job.name, "test_plugins.wait_10_seconds")
        self.assertEquals(self.job.uuid, "correctuuid")
        self.assertEquals(self.job._api_key, "api_key")

    def test_init_addr_does_not_exist_raises(self):
        self.assertRaises(
            ConnectionError,
            RemoteJob,
            "localhost.not.the.right.url",
            5001,
            "test_plugins.wait_10_seconds",
            "uuid",
            "api_key"
        )

    def test_init_job_name_does_not_exist_no_uuid_given_raises(self):
        with HTTMock(post_jobs_400):
            self.assertRaises(JobDoesNotExistError, RemoteJob, "localhost", 5001, "test_plugins.wait")

    def test_init_job_no_uuid_given_creates_job(self):
        with HTTMock(post_jobs, get_specific_job_named_something):
            self.job = RemoteJob("localhost", 5001, "something")
        self.assertEquals(self.job.uuid, "correctuuid")
        self.assertEquals(self.job.name, "something")

    def test_config(self):
        with HTTMock(get_specific_job):
            self.assertEquals(self.job.config(), job_dict['config'])

    def test_status(self):
        with HTTMock(get_specific_job):
            self.assertEquals(self.job.status(), job_dict['status'])

    def test_start(self):
        with HTTMock(get_specific_job, start_job):
            self.assertTrue(self.job.start())

    def test_kill(self):
        with HTTMock(get_specific_job, kill_job):
            self.assertTrue(self.job.kill())

    def test_running(self):
        with HTTMock(get_specific_job):
            self.assertTrue(self.job.running())
        with HTTMock(get_specific_job_running_false_finished_true):
            self.assertFalse(self.job.running(force=True))

    def test_finished(self):
        with HTTMock(get_specific_job):
            self.assertFalse(self.job.finished())
        with HTTMock(get_specific_job_running_false_finished_true):
            self.assertTrue(self.job.finished(force=True))

    def test_join(self):
        with HTTMock(get_specific_job_running_false_finished_true):
            self.assertTrue(self.job.join())

    def test_join_raises_timeouteror(self):
        with HTTMock(get_specific_job_named_something):
            self.assertRaises(TimeoutError, self.job.join, 0)

    def test_rate_limit(self):
        with HTTMock(get_specific_job):
            self.assertTrue(self.job.running())
        with HTTMock(get_specific_job_running_false_finished_true):
            self.assertTrue(self.job.running())

    def test_exception_thrown_from_status(self):
        with HTTMock(get_with_exception):
            self.assertRaises(JobFailedError, self.job.status, True)

    def test_exception_thrown_from_join(self):
        with HTTMock(get_with_exception):
            self.assertRaises(JobFailedError, self.job.join)

    def test_exception_thrown_from_status(self):
        with HTTMock(get_with_bubbled_up_exception):
            try:
                self.job.status()
            except JobFailedError as e:
                exception_info = exception_bubble_up_job["status"]["exception"]
                self.assertEqual(e.addr, exception_info["address"])
                self.assertEqual(e.uuid, exception_info["uuid"])
                self.assertEqual(e.type_string, exception_info["type"])
                self.assertEqual(e.msg, exception_info["message"])





from unittest2 import TestCase
from .utils import FlaskTestCaseMixin, StatusCodeTestMixin
from hoplite.builtin_plugins.install_python_package_job import install_package_with_pip, uninstall_package_with_pip
from hoplite.api.root import reload_site_packages
from tests.paths import TEST_PACKAGE_PATH


class HopliteTestCase(TestCase):
    def setUp(self):
        """
        This imports the test_jobs_package, which is required by many tests
        to exercise different parts of hoplite

        __import__ returns a module object which can then be used to invoke
        the different sub-modules imported by name in the last argument. I am
        using this because we install the package at runtime with pip,
        so we have to import at runtime.

        This simplifies development where new test plugins need to be created
        to exercise code. It also allows us to have one dev_requirements file
        instead of one per OS.
        """
        # Monkey patch pip so it doesn't mess with logging. Otherwise, presence
        # of nose xunit logging handlers will
        # cause an error when pip tries to set logging things
        def blank_func(blank_arg):
            pass
        import pip.basecommand
        pip.basecommand.__dict__['logging_dictConfig'] = blank_func

        install_package_with_pip(TEST_PACKAGE_PATH)
        reload_site_packages()
        self.test_jobs_module = __import__('test_jobs_package', globals(), locals(), ['constants',
                                                                                      'throw_an_exception_job',
                                                                                      'wait_10_seconds_job',
                                                                                      'create_file_job',
                                                                                      'throw_job_failed_exception'], -1)

    def tearDown(self):
        uninstall_package_with_pip('test-jobs')
        reload_site_packages()


class HopliteAppTestCase(HopliteTestCase, FlaskTestCaseMixin, StatusCodeTestMixin):
    def _create_app(self):
        raise NotImplementedError

    def setUp(self):
        super(HopliteAppTestCase, self).setUp()
        self.app = self._create_app()
        self.client = self.app.test_client()
        self.app_context = self.app.app_context()
        self.app_context.push()

    def tearDown(self):
        super(HopliteAppTestCase, self).tearDown()
        self.app_context.pop()



from httmock import urlmatch, HTTMock, response, all_requests
import psutil
import sys
import time
import unittest2

from hoplite.builtin_plugins.constants import DOWNLOAD_NETWORK_FOLDER_JOB_NAME, DOWNLOAD_FOLDER_FROM_FTP_JOB_NAME
from hoplite.client.remote_job_manager import RemoteJobManager
from hoplite.exceptions import InternalServerError
from hoplite.public_api import wait_for_hoplite
from hoplite.serializer import hoplite_dumps
from hoplite.server.jobs.job_manager import JobDoesNotExistError
from hoplite.utils import server_logging

logger = server_logging.get_server_logger(__name__)


def start_hoplite_server(port_num):
    proc = psutil.Popen(
        '{} -c "import hoplite.main; hoplite.main.server_main([\'--port={}\'])"'.format(sys.executable, port_num)
    )
    wait_for_hoplite('localhost', port_num)
    return proc


def tear_down_hoplite(process):
    l = process.get_children()
    for child in l:
        child.terminate()
        child.wait()
    process.terminate()
    process.wait()
    time.sleep(.5)


@urlmatch(netloc="localhost:5001", path='\/jobs\/running')
def get_running_jobs(url, request):
    if request.method == 'GET':
        return response(200, hoplite_dumps({"jobs": [{"uuid": 111}]}))

@all_requests
def response_500(url, request):
    return response(500)

class TestRemoteDaemonManager(unittest2.TestCase):
    def setUp(self):
        self.proc = start_hoplite_server(5001)
        self.manager = RemoteJobManager("localhost", 5001)

    def tearDown(self):
        tear_down_hoplite(self.proc)

    def test_create_job(self):
        job = None
        job = self.manager.create_job(DOWNLOAD_NETWORK_FOLDER_JOB_NAME, {})
        self.assertIsNotNone(job)
        self.assertEquals(job.name, DOWNLOAD_NETWORK_FOLDER_JOB_NAME)
        self.assertEquals(job.running(), False)

    def test_get_job(self):
        job = self.manager.create_job(DOWNLOAD_NETWORK_FOLDER_JOB_NAME, {})
        r_job = self.manager.get_job(job.uuid)
        self.assertEquals(r_job.uuid, job.uuid)
        self.assertEquals(r_job.name, job.name)

    def test_get_job_raises(self):
        self.assertRaises(JobDoesNotExistError, self.manager.get_job, 393939)

    def test_get_job_plugins(self):
        job_plugins = self.manager.get_job_plugins()
        self.assertEquals(len(job_plugins), 7)
        actual = job_plugins
        self.assertIn(DOWNLOAD_FOLDER_FROM_FTP_JOB_NAME, actual)
        self.assertIn(DOWNLOAD_NETWORK_FOLDER_JOB_NAME, actual)

    def test_get_job_plugins_raises(self):
        with HTTMock(response_500):
            self.assertRaises(InternalServerError, self.manager.get_job_plugins)

    def test_get_running_jobs(self):
        jobs = self.manager.get_running_jobs()
        self.assertEqual(len(jobs), 0)
        with HTTMock(get_running_jobs):
            jobs = self.manager.get_running_jobs()
            self.assertEqual(len(jobs), 1)

    def test_reload_site_packages(self):
        done = self.manager.reload_site_packages()
        self.assertTrue(done)

    def test_is_manager_available(self):
        time.sleep(1)  # Give some time for job manager to connect to start
        self.assertTrue(self.manager.is_available())
        invalid_manager = RemoteJobManager("localhost", 48423)
        self.assertFalse(invalid_manager.is_available())


from hoplite.client.remote_job_manager import RemoteJobManager
from hoplite.builtin_plugins import constants


class Error(Exception):
    pass


def remote_install(host, package, port=5000, options=''):
    '''
    Provides a convient interface to install a python package
    through the hoplite job mechanism.

    :param string host: The host machine to install pip package
    :param string package: The package to install
    :param int port: The hoplite port to use (5000 by default)
    :param string options: The pip options to use ('' by default)
    '''
    job_manager = RemoteJobManager(host, port)
    job = job_manager.create_job(
        plugin_name=constants.INSTALL_PYTHON_PACKAGE_JOB_NAME,
        config={
            'PIP_CMD': 'install {0} {1}'.format(options, package)
        })
    job.start()
    job.join()
    status = job.status()
    if not status['succeeded']:
        raise Error(status['stdout'])



"""
Module for providing a customized json serializer for Hoplite to use.

Hoplite requires a json serializer that:
    -Can handle mongo related objects(objectID, datetime)
    -Has a decoder that can return dictionaries ASCII(UTF-8 to be specific)
        keywords and text
    -Does not add information to your dictionary when decoding that it thinks
        may be useful like bson.json_util does for datetime objects.

    bson.json_util does most of this, so that's what we use for the most part.
        We copy and modify the
    bson.json_util.loads function and object hook in this module to make it
        satisfy the requirements described above.
"""

import base64
import datetime
import json
import re
import uuid

import bson
from bson import EPOCH_AWARE, EPOCH_NAIVE
from bson.binary import Binary
from bson.code import Code
from bson.dbref import DBRef
from bson.int64 import Int64
from bson.max_key import MaxKey
from bson.min_key import MinKey
from bson.objectid import ObjectId
from bson.regex import Regex
from bson.timestamp import Timestamp


from bson.json_util import dumps

_RE_OPT_TABLE = {
    "i": re.I,
    "l": re.L,
    "m": re.M,
    "s": re.S,
    "u": re.U,
    "x": re.X,
}


def hoplite_dumps(obj, *args, **kwargs):
    """
    Serializes a dictionary into unicode(unless specified otherwise)
    bson.json_util.dumps does exactly what hoplite needs, so that's what
    we call
    :param obj: Python dictionary to be serialized
    :param args: Please refer to online documentation for bson.json_util.dumps
        and json.dumps
    :param kwargs: Please refer to online documentation for
        bson.json_util.dumps and json.dumps
    :return: serialized obj in unicode
    """
    return dumps(obj, *args, **kwargs)


def hoplite_loads(s, *args, **kwargs):
    """
    Decodes a serialized dictionary.
    :param s: serialized dictionary
    :return: unserialized dictionary
    """
    compile_re = kwargs.pop('compile_re', True)
    ensure_tzinfo = kwargs.pop('ensure_tzinfo', False)
    encoding = kwargs.pop('encoding', 'utf-8')
    kwargs['object_hook'] = lambda dct: object_hook(dct, compile_re, ensure_tzinfo, encoding)
    return json.loads(s, *args, **kwargs)


def object_hook(dct, compile_re=False, ensure_tzinfo=True, encoding=None):
    """
    Object hook used by hoplite_loads. This object hook can encode the
    dictionary in the right text format.  For example, json.loads by default
    will decode '{'hey':'hey'}' into {u'hey':u'hey'} rather than {'hey':'hey'}.
    If encoding is set to utf-8, this object_hook can make '{'hey':'hey'}'
    decode to {'hey':'hey'} This object hook also decodes extended json types
    such as objectId and datetime objects. Datetime objects also
    have the option to be decoded with or without timezone information.
    :param dct: Dictionary this object hook is to operate on.
    :param ensure_tzinfo: Boolean deciding if timezone info should be added to
        decoded datetime objects
    :param encoding: choice of text decoding(unicode/utf-8, perhaps others)
    :return:
    """

    if encoding:
        # Converts all keys and unicode values in the top layer of the current
        # dictionary to the desired encoding type.
        new_dct = {}
        for key, value in list(dct.items()):
            if isinstance(key, str):
                key = key.encode(encoding)
            if isinstance(value, str):
                value = value.encode(encoding)
            new_dct[key] = value
        dct = new_dct

    if "$oid" in dct:
        return ObjectId(str(dct["$oid"]))
    if "$ref" in dct:
        return DBRef(dct["$ref"], dct["$id"], dct.get("$db", None))
    if "$date" in dct:
        secs = float(dct["$date"]) / 1000.0
        if ensure_tzinfo:
            return EPOCH_AWARE + datetime.timedelta(seconds=secs)
        else:
            # Avoid adding time zone info by default, unlike
            # bson.json_util.loads. If the developer really wants this, they
            # will have to specify it.
            return EPOCH_NAIVE + datetime.timedelta(seconds=secs)

    if "$regex" in dct:
        flags = 0
        # PyMongo always adds $options but some other tools may not.
        for opt in dct.get("$options", ""):
            flags |= _RE_OPT_TABLE.get(opt, 0)
        if compile_re:
            return re.compile(dct["$regex"], flags)
        else:
            return Regex(dct["$regex"], flags)
    if "$minKey" in dct:
        return MinKey()
    if "$maxKey" in dct:
        return MaxKey()
    if "$binary" in dct:
        if isinstance(dct["$type"], int):
            dct["$type"] = "%02x" % dct["$type"]
        subtype = int(dct["$type"], 16)
        if subtype >= 0xffffff80:  # Handle mongoexport values
            subtype = int(dct["$type"][6:], 16)
        return Binary(base64.b64decode(dct["$binary"].encode()), subtype)
    if "$code" in dct:
        return Code(dct["$code"], dct.get("$scope"))
    if "$uuid" in dct:
        return uuid.UUID(dct["$uuid"])
    if "$undefined" in dct:
        return None
    if "$numberLong" in dct:
        return Int64(dct["$numberLong"])
    if "$timestamp" in dct:
        tsp = dct["$timestamp"]
        return Timestamp(tsp["t"], tsp["i"])
    return dct


if __name__ == '__main__':
    dictionary = {
                        "uuid": "correctuuid",
                        "name": "test_plugins.wait_10_seconds",
                        "config": {"something": "is missing"},
                        "status":
                        {
                            "Roger": "Roger what's your status?",
                            "anumber": 123,
                            "objectid": bson.objectid.ObjectId(),
                            "datetime": datetime.datetime.now(),
                            "regexp": re.compile("hoplite")
                        },
                        "running": True,
                        "finished": False

                }
    print((hoplite_dumps(dictionary)))
    dictionary_bson_back = hoplite_loads(
        hoplite_dumps(hoplite_loads(hoplite_dumps(dictionary))))
    print(dictionary)
    print(dictionary_bson_back)
    print(("Dictionaries are the same: " + str((
        dictionary == dictionary_bson_back))))



from hoplite.utils import server_logging
from hoplite.serializer import hoplite_loads
from flask import Blueprint, request
from hoplite.api.helpers import job_manager, jsonify
from hoplite.exceptions import (
    JobDoesNotExistError,
    JobPluginDoesNotExistError,
    JobNotStartedError,
    JobAlreadyStartedError)

logger = server_logging.get_server_logger(__name__)


bp = Blueprint('jobs', __name__)


@bp.route("", methods=['GET'])
def get_jobs():
    logger.debug(
        "HTTP: Request All Jobs - From: {0}".format(request.remote_addr))
    jobs = job_manager.all_jobs()
    jobs_as_dict = []
    for job in jobs:
        jobs_as_dict.append(job.to_dict())
    return jsonify(jobs=jobs_as_dict)


@bp.route("", methods=['POST'])
def create_job():
    job_dict = hoplite_loads(request.data)
    name = job_dict.get('name', "default")
    config = job_dict.get('config', {})
    running = job_dict.get('running', False)
    port = job_dict.get('port', 5000)
    try:
        logger.debug(
            "HTTP: Request Create Job:{0} - From: {1}".format(
                name, request.remote_addr))
        job = job_manager.create_job(name, config, running, port)
    except JobPluginDoesNotExistError as e:
        return jsonify(error=str(e)), 400
    return jsonify(**job.to_dict())


@bp.route("/<job_uuid_string>", methods=['GET', 'PUT'])
def created_job(job_uuid_string):
    logger.debug(
        "HTTP: {0} Status Job UUID:{1} - From: {2}".format(
            request.method, job_uuid_string, request.remote_addr))
    try:
        job = job_manager.get_job(job_uuid_string)
    except (JobDoesNotExistError, ValueError) as e:
        return jsonify(error=str(e)), 404
    if request.method == 'PUT':
        r_json = hoplite_loads(request.data)
        if r_json.get("status", None):
            job.update_status(r_json["api_key"], r_json["status"])
    return jsonify(**job.to_dict())


@bp.route("/<job_uuid>/start", methods=['PUT'])
def start_job(job_uuid):
    try:
        logger.debug(
            "HTTP: Start Job UUID:{0} - From: {1}".format(
                job_uuid, request.remote_addr))
        job = job_manager.get_job(job_uuid)
        job.start()
    except JobDoesNotExistError as e:
        return jsonify(error=str(e)), 404
    except JobAlreadyStartedError as e:
        return jsonify(error=str(e)), 403
    return jsonify(uuid=job.uuid, started=True)


@bp.route("/<job_uuid>/kill", methods=['PUT'])
def kill_job(job_uuid):
    logger.debug(
        "HTTP: Terminate Job UUID:{0} - From: {1}".format(
            job_uuid, request.remote_addr))
    try:
        job = job_manager.get_job(job_uuid)
        job.kill()
    except JobDoesNotExistError as e:
        return jsonify(error=str(e)), 404
    except JobNotStartedError as e:
        return jsonify(error=str(e)), 403
    return jsonify(uuid=job.uuid, killed=True)


@bp.route("/running")
def running_jobs():
    logger.debug(
        "HTTP: Get Running Jobs- From: {0}".format(request.remote_addr))
    running_jobs = []
    jobs = job_manager.all_jobs()
    for job in jobs:
        if job.running():
            running_jobs.append(job.to_dict())
    return jsonify(jobs=running_jobs)



import time

from hoplite.builtin_plugins.constants import DOWNLOAD_NETWORK_FOLDER_JOB_NAME
from hoplite.utils import server_logging
from hoplite.server.jobs.job import Job, JobNotStartedError, NotAuthorizedError
from tests import HopliteTestCase

logger = server_logging.get_server_logger(__name__)


class TestJob(HopliteTestCase):
    def setUp(self):
        super(TestJob, self).setUp()
        self.job = Job("{3939}", DOWNLOAD_NETWORK_FOLDER_JOB_NAME, { "path": "/path/to/something" }, "temp")

    def test_initialize(self):
        self.assertEquals(self.job.uuid, "{3939}")
        self.assertEquals(self.job.name, DOWNLOAD_NETWORK_FOLDER_JOB_NAME)
        self.assertEquals(self.job.config, { "path": "/path/to/something" })
        self.assertEquals(self.job.running(), False)

    def test_start(self):
        self.job.start()
        self.assertTrue(self.job.running())
        self.job.kill()

    def test_finished_throws_if_job_not_started(self):
        with self.assertRaises(JobNotStartedError):
            self.job.finished()

    def test_finished(self):
        job = Job("No ID", self.test_jobs_module.constants.THROW_AN_EXCEPTION_JOB_NAME, { "No": "Config" }, "some_complex_key")
        job.start()
        while job.running():
            pass
        self.assertTrue(job.finished())

    def test_to_dict(self):
        d = self.job.to_dict()
        self.assertEquals(d["uuid"], "{3939}")
        self.assertEquals(d["name"], DOWNLOAD_NETWORK_FOLDER_JOB_NAME)
        self.assertEquals(d["config"], { "path": "/path/to/something" })
        self.assertEquals(d["status"], {})
        self.assertFalse(d["running"])
        self.assertFalse(d["finished"])

    def test_returns_exception_information_in_status(self):
        config = {}
        job = Job("666", self.test_jobs_module.constants.THROW_AN_EXCEPTION_JOB_NAME, config, "api_key", entry_point_group_name='hoplite.test_jobs')
        job.start()
        while job.running():
            time.sleep(.01)
        exc_info = job.status()["exception"]
        traceback = None
        # Get to the bottom level of the exception information
        while ('type' not in exc_info) and (exc_info is not None):
            traceback = exc_info.get("traceback", None)
            exc_info = exc_info.get('previous_exception', None)
        self.maxDiff = None
        self.assertEqual(exc_info["type"], str(TypeError))
        self.assertEqual(exc_info["message"], "THE SKY IS FALLING!!")
        self.assertIsNotNone(traceback)

    def test_returns_status(self):
        self.assertEqual(self.job.status(), {})

    def test_updates_status(self):
        self.job.update_status(self.job._api_key, { "slave_ip": "10.2.13.123" })
        status = self.job.status()
        self.assertEquals(status, { "slave_ip": "10.2.13.123" })
        # Make sure it overwrites old key values
        self.job.update_status(self.job._api_key, { "slave_ip": "12.3.4.567" })
        status = self.job.status()
        self.assertEquals(status, { "slave_ip": "12.3.4.567" })

    def test_update_status_raises_on_invalid_api_key(self):
        self.assertRaises(NotAuthorizedError, self.job.update_status, "", {"Not": "Authorized"})

    def test_kill(self):
        job = Job("No ID", self.test_jobs_module.constants.WAIT_10_SECONDS_JOB_NAME, { "No": "Config" }, "temp_api_key")
        logger.info("Starting job")
        job.start()
        self.assertTrue(job.running())
        job.kill()
        start_time = time.time()
        while job.running():
            if time.time() - start_time > 1:
                raise Exception("Job not killed in time")
        self.assertTrue(job.finished())



from flask import Flask
from hoplite.api.root import bp as site_bp
from hoplite.api.jobs import bp as jobs_bp
from hoplite.api.job_plugins import bp as job_plugins_bp
from hoplite.server.jobs.job_manager import JobManager
from hoplite.plugin_manager import EntryPointManager
import hoplite.api.helpers


def create_app(group_name='hoplite.jobs'):
    app = Flask(__name__)
    app.register_blueprint(site_bp, url_prefix='/')
    app.register_blueprint(jobs_bp, url_prefix='/jobs')
    app.register_blueprint(job_plugins_bp, url_prefix='/job_plugins')
    hoplite.api.helpers.manager = JobManager(EntryPointManager(group_name))
    return app



from hoplite.serializer import hoplite_dumps
from hoplite.exceptions import InternalServerError
import requests


class ClientMixin(object):
    def _json_data(self, kwargs):
        if 'data' in kwargs:
            kwargs['data'] = hoplite_dumps(kwargs['data'])
        if not kwargs.get('headers', None):
            kwargs['headers'] = {'Content-type': 'application/json'}
        return kwargs

    def _raise_if_status_500(self, response):
        if response.status_code == 500:
            raise InternalServerError()
        return response

    def _request(self, method, *args, **kwargs):
        return self._raise_if_status_500(method(*args, **kwargs))

    def _jrequest(self, *args, **kwargs):
        return self._request(*args, **kwargs)

    def jget(self, *args, **kwargs):
        return self._jrequest(requests.get, *args, **kwargs)

    def jpost(self, *args, **kwargs):
        return self._jrequest(requests.post, *args, **self._json_data(kwargs))

    def jput(self, *args, **kwargs):
        return self._jrequest(requests.put, *args, **self._json_data(kwargs))

    def jpatch(self, *args, **kwargs):
        return self._jrequest(requests.patch, *args, **self._json_data(kwargs))

    def jdelete(self, *args, **kwargs):
        return self._jrequest(requests.delete, *args, **self._json_data(kwargs))



"""
Installs a python package to the local python environment (directory containing
a setup.py)

Site-packages is reloaded after the install, so the package will be available
to import and the entry points will be registered correctly.

Example config::

    {
        KEYS.LOCAL_PATH: "C:\\path\\to\\package\\folder"
    }

Example return status::

    No Error:
    {
        "succeeded": True,
        "stdout": "All the stdout output"
    }

    Error (pip install failed):

    {
        "succeeded": False,
        "stdout": "All stdout",
        "errors": ["Error string"]
    }
"""
from hoplite.builtin_plugins.constants import (
    InstallPythonPackageJobConstants as KEYS)
from hoplite.client.remote_job_manager import RemoteJobManager
import pip
from io import StringIO
import sys


def install_package_with_pip(pip_cmd):
    out = None
    if pip_cmd is not None:
        backup = sys.stdout
        string_io = StringIO()
        sys.stdout = string_io
        pip_cmd = pip_cmd.split()
        if 'install' not in pip_cmd:
            pip_cmd.insert(0, 'install')
        ret = pip.main(pip_cmd)
        sys.stdout = backup
        out = string_io.getvalue()
        # FIXME: We should be calling string_io.close() but pip errors when we
        # do that
    return ret, out


def uninstall_package_with_pip(package_name):
    pip.main(['uninstall', package_name, '-y'])


def run(config, status):
    manager = RemoteJobManager("localhost", 5000)
    pip_cmd = config.get(KEYS.PIP_CMD, None) if config.get(KEYS.PIP_CMD, None) is not None else config.get(KEYS.LOCAL_PATH, None)

    if pip_cmd is None:
        status.update(
            {"succeeded": False, "errors": ["No local path specified"]})
        return
    ret, stdout = install_package_with_pip(pip_cmd)
    if ret:
        status.update(
            {"succeeded": False, "stdout": stdout, "errors": ["Pip returned a non-zero error code"]})
        return
    manager.reload_site_packages()
    status.update({"succeeded": True, "stdout": stdout})



from hoplite.utils import server_logging
from flask import Blueprint, request
from hoplite.api.helpers import job_manager, jsonify

logger = server_logging.get_server_logger(__name__)

bp = Blueprint("job_plugins", __name__)


@bp.route("")
def available_job_plugins():
    logger.debug(
        "HTTP: Get Available Job Plugins- From: {0}".format(
            request.remote_addr))
    return jsonify(job_plugins=job_manager.available_job_plugins())



from multiprocessing import Process
from multiprocessing import Pipe

from hoplite.utils import server_logging
from hoplite.exceptions import (
    JobAlreadyStartedError,
    JobNotStartedError,
    NotAuthorizedError)
from .job_wrapper import job_wrapper
from hoplite.client.status_updater import StatusUpdater
from hoplite.plugin_manager import EntryPointManager


class Job(object):
    """
    Represents a job that has been created on the server
    """
    def __init__(self, job_uuid, name, config, api_key, entry_point_group_name="hoplite.jobs", port=5000):
        """
        @param job_uuid unique identifier for this job
        @param name the name of the job, corresponds to the plugin name
        @param config dictionary object containing configuration for the
            specific job
        """
        self.port = port
        self.uuid = job_uuid
        self.name = name
        self.config = config
        self._api_key = api_key
        self._status = {}
        self._process = None
        self._started = False
        self._killed = False
        self._pipe_to_self = None
        self._pipe_to_process = None
        # TODO: We need this workaround because in tests I create jobs that
        # don't have a corresponding loaded entry point
        # At some point the tests should be refactored to use jobs that exist
        # and we can get rid of this code
        module = EntryPointManager().get_plugin_module_by_name(name)
        logger_name = module.__name__ if module is not None else name
        self._logger = server_logging.get_job_logger(
            logger_name, uuid=self.uuid)
        self._entry_point_group_name = entry_point_group_name

    def running(self):
        """
        The current state of the process executing the job
        :return: Boolean describing if job is running
        """
        if self._process:
            return self._process.is_alive()
        return False

    def killed(self):
        """
        Checked if job has been killed.
        :return: Boolean describing if job was killed.
        """
        return self._killed

    def start(self):
        """
        Start the job. If the job has already been started before, a
        JobAlreadyStartedError is raised.
        """
        if self._started:
            raise JobAlreadyStartedError(self.uuid)
        if self.name:
            # TODO Testing hole here...We don't make sure updater is good or
            # anything
            updater = StatusUpdater(
                'localhost:{}'.format(self.port), self.uuid, self._api_key)
            self._pipe_to_process, self._pipe_to_self = Pipe()
            self._logger.debug(
                "Starting Job {0} UUID:{1}".format(self.name, self.uuid))
            self._process = Process(
                target=job_wrapper,
                args=(
                    self._pipe_to_self,
                    self.name,
                    self.config,
                    updater,
                    self._entry_point_group_name,
                    self.uuid))
            self._process.start()
            self._started = True

    def finished(self):
        """
        Returns True once the job has finished running.
        Will raise a JobNotStartedError if :meth:`job.start` has not been
        called
        :return: Boolean describing if the process has started and is no longer
            running
        """
        if self._process is None:
            raise JobNotStartedError(self.uuid)
        return not self._process.is_alive() and self._started

    def status(self):
        """
        Update exception dictionary of status if there is something available
        from the process pipe. Returns the updated status afterwards.
        :return: status dictionary from the job processes
        """
        if self._pipe_to_process:
            if self._pipe_to_process.poll():
                exception_dictionary = self._pipe_to_process.recv()
                self._status["exception"] = exception_dictionary
                self._pipe_to_process = None
        return self._status

    def update_status(self, api_key, status_update):
        if api_key != self._api_key:
            raise NotAuthorizedError
        self._status = dict(list(self._status.items()) + list(status_update.items()))
        self._logger.debug(
            "Update Status:{0} UUID:{1} Status:{2}".format(
                self.name, self.uuid, self._status))

    def kill(self):
        """
        Kills the job if has been started. Raises a JobNotStartedError if it
        has not started yet. If the job has already finished, the job will
        still be flagged as killed, but it will not have any further
        consequences.
        """
        if self._process is None:
            raise JobNotStartedError(self.uuid)
        self._logger.debug(
            "Terminating Job:{0} UUID:{1}".format(self.name, self.uuid))
        self._process.terminate()
        self._pipe_to_process = None
        self._pipe_to_self = None
        self._killed = True

    def to_dict(self):
        """
        Returns a dictionary representation of the job.
        Used to serialize job data using JSON for sending over the network.
        """
        d = {}
        d["uuid"] = self.uuid
        d["name"] = self.name
        d["config"] = self.config
        d["status"] = self.status()
        d["running"] = self.running()
        d["killed"] = self.killed()
        try:
            d["finished"] = self.finished()
        except JobNotStartedError:
            d["finished"] = False
        return d



from hoplite.utils import server_logging
import os
import sys
import platform
from tests.api import HopliteApiTestCase
import tempfile
import shutil
import pip
import pkg_resources


logger = server_logging.get_server_logger(__name__)

class RootSiteTestCase(HopliteApiTestCase):
    def setUp(self):
        super(RootSiteTestCase, self).setUp()

    '''
    This test is failing intermittently. I think it has to do with how exactly things are installed into the
    virtual environment and when/how the packages are reloaded.

    def test_reload_site_packages(self):
        opsys = platform.system().lower()
        python_exe_path = sys.executable
        bin_path = os.path.split(python_exe_path)[0]
        venv_path = bin_path
        site_path = ""
        if opsys == 'windows':
            if hasattr(sys, 'real_prefix'):  # If running in virtual environment, we must account for different folder structure
                site_path = os.path.join(venv_path, "..", "Lib", "site-packages")
            else:
                site_path = os.path.join(venv_path, "Lib", "site-packages")
        elif opsys == 'linux' or opsys == 'darwin':
            if hasattr(sys, 'real_prefix'):  # If running in virtual environment, we must account for different folder structure
                site_path = os.path.join(venv_path, "..", "lib", "python{0}.{1}".format(sys.version_info[0], sys.version_info[1]),
                                        "site-packages")
            else:
                site_path = os.path.join(venv_path, "lib", "python{0}.{1}".format(sys.version_info[0], sys.version_info[1]),
                                        "site-packages")

        pth_file_path = os.path.join(site_path, "test.pth")
        try:
            pth_file = open(pth_file_path, 'w+')
            pth_file.write(os.path.abspath("./tests"))
            pth_file.close()
            r = self.client.put('/reload')
            self.assertOk(r)
        except Exception, e:
            os.remove(pth_file_path)
            raise e
        # This will raise if the path is not in sys.path
        sys.path.remove(os.path.abspath("./tests"))
    '''

    def test_loads_new_entry_points(self):
        setup_str = "from setuptools import setup, find_packages;setup(name='TestEntryPointPackage', version='0.1', packages=find_packages(), entry_points={'test.entry': ['job1=tepp.myjob']})"

        tempdir = tempfile.mkdtemp()
        try:
            setup_py = open(os.path.join(tempdir, "setup.py"), 'w')
            setup_py.write(setup_str)
            setup_py.close()
            package_path = os.path.join(tempdir, "tepp")
            os.mkdir(package_path)
            init_file = open(os.path.join(package_path, "__init__.py"), 'w')
            init_file.close()
            myjob = open(os.path.join(package_path, "myjob.py"), 'w')
            myjob.close()

            pip.main(['install', tempdir])
            r = self.client.put('/reload')
            self.assertOk(r)

            count = 0
            for entry_point in pkg_resources.iter_entry_points(group="test.entry"):
                self.assertEqual(entry_point.name, "job1")
                count += 1
                if count > 1:
                    self.fail("test.entry entry point has more than one definition")
        finally:
            pip.main(['uninstall', '-y', "TestEntryPointPackage"])
            shutil.rmtree(tempdir)





import logging
import pickle
import time
import traceback

from hoplite.utils.server_logging import add_remote_function_logging_handlers

hoplite_logger = logging.getLogger()
hoplite_logger.setLevel(logging.DEBUG)


def run(config, status):
    """
    Run a function called from another machine.

    :param config: Dictionary which contains information needed to run the
        function.
        It must contain the following keys:
        * args - The original non-keyword arguments which are intended to be
          sent to the function being called
        * kwargs - The original keyword arguments which are intended to be sent
          to the function being called
        * module_name - The name of the module which defines the class on which
          the function will be called. This module will be imported so that the
          class can be unpickled and then used.
        * function_name - The name of the function to call from the class
          object.
        * instance - The pickled instance of the class from which the function
          will be called
    :param status: Used to store information regarding the results of running
        the job
    """
    args = pickle.loads(config.get('args', []))
    kwargs = pickle.loads(config.get('kwargs', {}))
    self = pickle.loads(config.get('instance'))
    function_name = config.get('function_name')
    status.status.update({
        'Positional arguments for function call': str(args),
        'Keyword arguments for function call': str(kwargs)
    })
    status.update(status.status)
    try:
        # Attempt to get namespaced class name for use in creating log
        # hierarchy
        try:
            class_name = str(self.__class__)
            if class_name.startswith("<class '"):
                class_name = class_name[8:]
            if class_name.endswith("'>"):
                class_name = class_name[:len(class_name) - 2]
        except Exception:
            class_name = str(self.__class__.__name__)
        timestamp = time.strftime('%Y-%m-%d %H-%M-%S')
        log_filename = add_remote_function_logging_handlers(
            class_name + '.' + function_name, timestamp)
        hoplite_logger.info(
            'Beginning execution of {} with args: {} and kwargs: {}'.format(
                function_name, args, kwargs))
        # Run function
        return_values = getattr(self, function_name)(*args, **kwargs)
        # Remove and re-add handlers because the function could have messed
        # with the root logger
        all_handlers = list(hoplite_logger.handlers)
        for handler in all_handlers:
            handler.close()
            hoplite_logger.removeHandler(handler)
        # We need to pass in log filename this time so that a new file isn't
        # created
        add_remote_function_logging_handlers(
            class_name + '.' + function_name, timestamp, log_filename)
    except:
        hoplite_logger.error(
            'An exception occurred: \n{}'.format(
                traceback.format_exc().rstrip()))
        status.status.update({
            'Exception logging info': traceback.format_exc().replace('\\n', '\n')
        })
        status.update(status.status)
        raise

    hoplite_logger.info(
        'Returning from {} with return value(s): {}'.format(
            function_name, return_values))

    if return_values.__class__ is tuple:
        return_values = list(return_values)
    else:
        return_values = [return_values]

    status.status.update({
        'return_values': pickle.dumps(return_values)
    })
    status.update(status.status)

    for handler in all_handlers:
        handler.close()
        hoplite_logger.removeHandler(handler)



from hoplite.builtin_plugins.utils.network_file_operations import upload_folder_to_network_directory
from hoplite.client.status_updater import MockStatusUpdater


def run(config, status):
    """
    Uploads a folder from the local disk to a network location
    This job is for Windows only (you better remember to escape your
    backslashes)
    **dest_path must not exist**
    Expected Config:
    {
        network_path: "\\\\share\\something\\something\\something",
        local_path: "C:\\Documents\\doesnotexist",
        user: "DOMAIN\\username",
        password: "password"
    }
    """
    network_path = config.get("network_path", None)
    local_path = config.get("local_path", None)
    user = config.get("user", "")
    password = config.get("password", "")

    if network_path is None:
        status.update({"error": "network_path not defined in config"})
        return
    if local_path is None:
        status.update({"error": "local_path not defined in config"})
        return

    if network_path.find('\\\\', 0, 2) == 0:
        stdout = ""
        try:
            stdout = upload_folder_to_network_directory(
                network_path, local_path, user, password)
        except RuntimeError as e:
            status.update({"error": e})
        status.update({"success": True, "stdout": stdout})
        return
    status.update({"error": "Invalid network path"})



import logging
import time

from exceptions import ExternalCustomError, ExternalEmptyError
from hoplite.remote_enabler import remotify

logger = logging.getLogger(__name__)


@remotify(__name__)
def return_none():
    """
    Used to test that a None value can be returned properly
    :return: None
    """
    None


@remotify(__name__)
def return_number():
    """
    Used to test that a single value can be returned properly
    :return: A number, 777
    """
    return 777


@remotify(__name__)
def return_tuple():
    """
    Used to test that a tuple can be returned properly
    :return: A tuple containing a string, an int, and a float
    """
    return 'All', 4, 1.11


@remotify(__name__)
def return_single_list():
    """
    Used to test that a single list can be returned properly
    :return: A list containing several strings
    """
    return ['This', 'is', 'a', 'list']


@remotify(__name__)
def return_multiple_lists():
    """
    Used to test that a jagged, 2-dimensional list can be returned properly
    :return: A list of lists, each of which are different lengths and contain
        different values
    """
    return ['This', 'is', 1, 'list'], ['and', 'this', 'is', 'another', 'list']


@remotify(__name__)
def do_math(number_1, number_2):
    """
    Used to test that numeric arguments can be passed through as arguments
    :return:
    """
    return number_1 * number_2


@remotify(__name__)
def pass_common_class(date):
        """
        Used to test that more complex class objects can be passed through as
        arguments
        :param date: Datetime object
        :return: A tuple containing the information stored in the datetime
        argument
        """
        return date.year, date.month, date.day, date.hour, date.minute, date.second, date.microsecond


@remotify(__name__)
def pass_custom_class(dummy):
        """
        Used to test that custom class objects can be passed through as'
        arguments
        :param dummy: DummyClass
        :return: The sum of numbers contained in the class
        """
        return dummy.add_numbers()


@remotify(__name__)
def raise_type_error():
    """
    Test that normal exceptions can be raised properly
    :raise TypeError: Unconditionally raises Type Error
    """
    raise TypeError('___Failure1')


@remotify(__name__)
def raise_custom_error():
    """
    Used to test that custom exceptions can be raised properly
    :raise ExternalEmptyError: Unconditionally raises ExternalEmptyError
    """
    raise ExternalEmptyError('___Failure2')


@remotify(__name__)
def create_specified_file(filename, contents='default_contents'):
    """
    Used to test remote file creation using a string passed into the function.
    The filename is also passed in. This is specifically used to test the
    asynchronous functionality of the metaclass.
    """
    time.sleep(1)
    with open(filename, 'w') as fout:
        fout.write(contents)


@remotify(__name__)
def call_nested_function():
    """
    Used to test that exception handling works properly when jobs are nested.
    This function starts a job that raises a type error.
    :raise TypeError: Raised when the function remote_raise_type_error is
        called
    """
    remote_raise_type_error('localhost:5001')


@remotify(__name__)
def raise_private_error():
    """
    Used to test that private errors defined within a class cannot be properly
    pickled. This means that the exception should come out on the client side
    as a JobFailedError instead.
    :raise JobFailedError: Attempts to raise MyPrivateError, but will fail and
        instead raise JobFailedError
    """
    class MyPrivateError(Exception):
        pass
    raise MyPrivateError('___Failure3')


@remotify(__name__)
def raise_public_error():
    """
    Used to test that externally defined, custom errors can be raised on the
    client side.
    :raise ExternalCustomError: Unconditionally raises ExternalCustomError
    """
    raise ExternalCustomError('___Failure4')


@remotify(__name__)
def return_custom_exception():
    """
    Used to test that custom class instances (such as exceptions) can be
    returned
    :return: Returns an ExternalCustomError exception object
    """
    return ExternalCustomError('___Failure4')


@remotify(__name__)
def long_job():
    """
    Used to test job timeouts on remotely called functions
    """
    time.sleep(7)


@remotify(__name__)
def log_normal(dummy_var_1, dummy_var_2=None):
    logger.info('Logging in log_normal function')
    return str(dummy_var_1) + ' + ' + str(dummy_var_2)


@remotify(__name__)
def log_nested_caller():
    logger.info('Currently in caller function')
    log_nested_callee()


def log_nested_callee():
    logger.info('Currently in callee function')



from flask import Blueprint
import platform
import sys
import os
import site
import pkg_resources

from hoplite.api.helpers import jsonify

bp = Blueprint('site', __name__)


def reload_site_packages():
    # TODO: Can probably be replaced with reload(pkg_resources)
    opsys = platform.system().lower()
    python_exe_path = sys.executable
    bin_path = os.path.split(python_exe_path)[0]
    venv_path = os.path.split(bin_path)[0]
    site_path = ""
    if opsys == 'windows':
        # If running in virtual environment, we must account for different
        # folder structure
        if hasattr(sys, 'real_prefix'):
            site_path = os.path.join(venv_path, "..", "Lib", "site-packages")
        else:
            site_path = os.path.join(venv_path, "Lib", "site-packages")
    elif opsys == 'linux' or opsys == 'darwin':
        # If running in virtual environment, we must account for different
        # folder structure
        if hasattr(sys, 'real_prefix'):
            site_path = os.path.join(
                venv_path, "..", "lib", "python{0}.{1}".format(
                    sys.version_info[0], sys.version_info[1]), "site-packages")
        else:
            site_path = os.path.join(
                venv_path, "lib", "python{0}.{1}".format(
                    sys.version_info[0], sys.version_info[1]), "site-packages")
    site.addsitedir(site_path)
    for path in sys.path:
        pkg_resources.working_set.add_entry(path)


@bp.route("reload", methods=["PUT"])
def reload_site_packages_endpoint():
    reload_site_packages()
    return jsonify()



import psutil
import subprocess
import sys
import time
import unittest2

from hoplite.public_api import wait_for_hoplite
import remotable_test_resources.remotable_class as remotable_class
import remotable_test_resources.remotable_module as remotable_module


def start_hoplite_server(port_num):
    proc = psutil.Popen(
        '{} -c "import hoplite.main; hoplite.main.server_main([\'--port={}\'])"'.format(sys.executable, port_num),
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT
    )
    wait_for_hoplite('localhost', port_num)
    return proc


def tear_down_hoplite(process):
    l = process.get_children()
    for child in l:
        child.terminate()
        child.wait()
    process.terminate()
    process.wait()
    time.sleep(.5)


class TestRemotableClassMultipleHopliteInstance(unittest2.TestCase):
    def setUp(self):
        """
        This function has to use subprocess instead of multiprocessing. Otherwise, tests fail
        with the error "daemonic processes are not allowed to have children". This way, hoplite is
        run in a separate process.
        :return:
        """
        self.proc1 = start_hoplite_server(5001)
        self.proc2 = start_hoplite_server(5002)
        self.proc3 = start_hoplite_server(5003)
        self.proc4 = start_hoplite_server(5004)
        self.class_instance = remotable_class.TestClass('this_is@a#string', 12349)

    def tearDown(self):
        tear_down_hoplite(self.proc1)
        tear_down_hoplite(self.proc2)
        tear_down_hoplite(self.proc3)
        tear_down_hoplite(self.proc4)

    def test_return_none(self):
        ret1 = self.class_instance.remote_return_none('localhost:5001')
        ret2 = self.class_instance.remote_return_none('localhost:5002')
        ret3 = self.class_instance.remote_return_none('localhost:5003')
        ret4 = self.class_instance.remote_return_none('localhost:5004')
        self.assertIsNone(ret1)
        self.assertIsNone(ret2)
        self.assertIsNone(ret3)
        self.assertIsNone(ret4)

    def test_do_math(self):
        ret1 = self.class_instance.remote_do_math('localhost:5001', 15, 20)
        ret2 = self.class_instance.remote_do_math('localhost:5002', 7, 5)
        ret3 = self.class_instance.remote_do_math('localhost:5003', 2, 3)
        ret4 = self.class_instance.remote_do_math('localhost:5004', 133, 289)
        self.assertEqual(ret1, 15 * 20)
        self.assertEqual(ret2, 7 * 5)
        self.assertEqual(ret3, 2 * 3)
        self.assertEqual(ret4, 133 * 289)

    def test_do_math_async(self):
        job1 = self.class_instance.remote_async_do_math('localhost:5001', 15, 20)
        job2 = self.class_instance.remote_async_do_math('localhost:5002', 7, 5)
        job3 = self.class_instance.remote_async_do_math('localhost:5003', 2, 3)
        job4 = self.class_instance.remote_async_do_math('localhost:5004', 133, 289)
        job1.start()
        job2.start()
        job3.start()
        job4.start()
        while not job1.finished() and not job2.finished() and not job3.finished() and not job4.finished():
            time.sleep(.25)
        ret1 = job1.join()
        ret2 = job2.join()
        ret3 = job3.join()
        ret4 = job4.join()
        self.assertEqual(ret1, 15 * 20)
        self.assertEqual(ret2, 7 * 5)
        self.assertEqual(ret3, 2 * 3)
        self.assertEqual(ret4, 133 * 289)


class TestRemotableModuleMultipleHopliteInstance(unittest2.TestCase):
    def setUp(self):
        """
        This function has to use subprocess instead of multiprocessing. Otherwise, tests fail
        with the error "daemonic processes are not allowed to have children". This way, hoplite is
        run in a separate process.
        :return:
        """
        self.proc1 = start_hoplite_server(5001)
        self.proc2 = start_hoplite_server(5002)
        self.proc3 = start_hoplite_server(5003)
        self.proc4 = start_hoplite_server(5004)

    def tearDown(self):
        tear_down_hoplite(self.proc1)
        tear_down_hoplite(self.proc2)
        tear_down_hoplite(self.proc3)
        tear_down_hoplite(self.proc4)

    def test_return_none(self):
        ret1 = remotable_module.remote_return_none('localhost:5001')
        ret2 = remotable_module.remote_return_none('localhost:5002')
        ret3 = remotable_module.remote_return_none('localhost:5003')
        ret4 = remotable_module.remote_return_none('localhost:5004')
        self.assertIsNone(ret1)
        self.assertIsNone(ret2)
        self.assertIsNone(ret3)
        self.assertIsNone(ret4)

    def test_do_math(self):
        ret1 = remotable_module.remote_do_math('localhost:5001', 15, 20)
        ret2 = remotable_module.remote_do_math('localhost:5002', 7, 5)
        ret3 = remotable_module.remote_do_math('localhost:5003', 2, 3)
        ret4 = remotable_module.remote_do_math('localhost:5004', 133, 289)
        self.assertEqual(ret1, 15 * 20)
        self.assertEqual(ret2, 7 * 5)
        self.assertEqual(ret3, 2 * 3)
        self.assertEqual(ret4, 133 * 289)

    def test_do_math_async(self):
        job1 = remotable_module.remote_async_do_math('localhost:5001', 15, 20)
        job2 = remotable_module.remote_async_do_math('localhost:5002', 7, 5)
        job3 = remotable_module.remote_async_do_math('localhost:5003', 2, 3)
        job4 = remotable_module.remote_async_do_math('localhost:5004', 133, 289)
        job1.start()
        job2.start()
        job3.start()
        job4.start()
        while not job1.finished() and not job2.finished() and not job3.finished() and not job4.finished():
            time.sleep(.25)
        ret1 = job1.join()
        ret2 = job2.join()
        ret3 = job3.join()
        ret4 = job4.join()
        self.assertEqual(ret1, 15 * 20)
        self.assertEqual(ret2, 7 * 5)
        self.assertEqual(ret3, 2 * 3)
        self.assertEqual(ret4, 133 * 289)


def tear_down_hoplite(process):
    l = process.get_children()
    for child in l:
        child.terminate()
        child.wait()
    process.terminate()
    process.wait()



import pkg_resources

from hoplite.utils import server_logging
import importlib
logger = server_logging.get_server_logger(__name__)


class EntryPointManager(object):
    """
    Interface to the hoplite.jobs set of entry points
    """
    def __init__(self, group_name='hoplite.jobs'):
        self.entry_point_group_name = group_name

    def get_plugin_names(self):
        names = []
        for entry_point in self._entry_points():
            names.append(entry_point.name)
        return names

    def get_plugin_module_by_name(self, name):
        # TODO: Should we validate the module implements the correct methods
        # before returning it to the caller?
        for entry_point in self._entry_points():
            if entry_point.name == name:
                return entry_point.load()

    def _entry_points(self):
        # TODO: If it takes too long to reload pkg_resources, it might be worth
        # looking into only loading if a package
        # TODO: isn't found and then try again.
        importlib.reload(pkg_resources)  # makes sure entry points are up-to-date
        for entry_point in pkg_resources.iter_entry_points(group=self.entry_point_group_name):
            logger.debug("Found entry point: {0}".format(entry_point.name))
            yield entry_point



from functools import wraps
import logging
import pickle
import sys
import types

from hoplite import client
from hoplite.exceptions import JobFailedError
from .globals import HopliteClientSettings


def remotify(module_name, functions=None, add_documentation=True):
    """
    Decorator which can be used to add remote capabilities to functions or
    classes.

    By using this decorator on a function, the module which contains the
    function will be given two additional functions which provide the means
    for calling the function on a remote machine running Hoplite. The names of
    these two functions are remote\_(name of original function) and
    remote_async\_(name of original function). The remote\_function calls the
    function or method asynchronously on the remote machine, and the
    remote_async\_ function returns an object which can be used to run the
    function asynchronously on the remote machine.

    By using this decorator on a class, the class will be enhanced with
    remote\_ and remote_async\_ methods which serve the same purpose as the
    remote\_ and remote_async\_ functions described above.

    In either case, documentation will be added to the new functions/methods
    giving a short description of how to use them and a link to the original
    function. This feature can be disabled by passing in False for the
    add_documentation parameter.

    For additional details, see the above documentation and examples.

    :param module_name: The namespace for the module which contains this class
      or function. This should simply be passed in as __name__
      (e.g. @remotify(__name__)
    :type module_name: str
    :param functions: A list of names of the methods which should be made
      remotable. This only applies when decorating classes. If no list is
      provided, then all methods in the class (besides those starting with __)
      will be made remotable
    :param add_documentation: If true, then the functions or methods which are
      added will be given documentation explaining how they are to be used, and
      linking to the original function. If false, then they will be given no
      documentation, though they will still show up in the Sphinx documentation
      if the associated rst file indicates that undocumented functions/methods
      should be included
    :type add_documentation: bool
    :type functions: list of [str]
    """
    if functions is None:
        functions = []

    def inner(obj):
        # If decorating a class, then add methods to the class. For some
        # reason, if a class which inherits from another class is decorated,
        # then the type of object passed in is "type" rather than "class".
        # However, the class can still be decorated just fine.
        if isinstance(obj, type) or isinstance(obj, type):
            class_obj = obj  # Rename for clarity
            class_item_names = dir(class_obj)
            for item_name in class_item_names:
                if len(functions) > 0 and item_name not in functions:
                    continue
                if isinstance(getattr(class_obj, item_name), types.MethodType):
                    func = getattr(class_obj, item_name)
                    name = func.__name__
                    # Skip "private" functions as well as inherited functions
                    # that have already been remoted
                    if name.startswith('__') or hasattr(func, '___remoted_by_hoplite___') \
                            or hasattr(func, '___is_hoplite_remotable___'):
                        continue
                    if name.startswith('remote_') or name.startswith('async_'):
                        raise AttributeError(
                            'Unable to add remote capability to function {0}:'
                            ' function cannot begin with "remote_" or '
                            '"async_"'.format(name))
                    class_func = wraps(func)(remote_func_builder(name))
                    async_class_func = wraps(func)(
                        remote_async_func_builder(name))
                    class_func.__name__ = 'remote_' + class_func.__name__
                    async_class_func.__name__ = 'remote_async_' + \
                        async_class_func.__name__
                    if add_documentation:
                        class_func.__doc__ = _get_remote_docstring(
                            'meth', '{}.{}'.format(
                                module_name, class_obj.__name__), func.__name__
                        )
                        async_class_func.__doc__ = _get_remote_async_docstring(
                            'meth', '{}.{}'.format(
                                module_name, class_obj.__name__), func.__name__
                        )
                    else:
                        class_func.__doc__ = None
                        async_class_func.__doc__ = None
                    # Need to set attribute on __func__, which is the
                    # underlying function stored in the instancemethod This
                    # adds a tag to the function being remotified so it is not
                    # remoted twice if a remoted class is inherited
                    setattr(func.__func__, '___remoted_by_hoplite___', True)
                    # Set attribute to remotable functions for identification
                    setattr(class_func, '___is_hoplite_remotable___', True)
                    setattr(
                        async_class_func, '___is_hoplite_remotable___', True)
                    setattr(class_obj, 'remote_' + name, class_func)
                    setattr(
                        class_obj, 'remote_async_' + name, async_class_func)
        # If decorating a module function (not a class function)
        elif isinstance(obj, types.FunctionType):
            func = obj  # Rename for clarity
            module = sys.modules[module_name]
            name = func.__name__
            if name.startswith('remote_') or name.startswith('async_'):
                raise AttributeError(
                  'Unable to add remote capability to function {0}:'
                  ' function cannot begin with "remote_" or "async_"'.format(
                      name))
            mod_func = wraps(func)(
                remote_module_func_builder(name, module_name))
            async_mod_func = wraps(func)(
                remote_module_async_func_builder(name, module_name))
            mod_func.__name__ = 'remote_' + mod_func.__name__
            async_mod_func.__name__ = 'remote_async_' + async_mod_func.__name__
            if add_documentation:
                mod_func.__doc__ = _get_remote_docstring(
                    'func', module_name, func.__name__)
                async_mod_func.__doc__ = _get_remote_async_docstring(
                    'func', module_name, func.__name__)
            else:
                mod_func.__doc__ = None
                async_mod_func.__doc__ = None
            # Set attribute to remotable and remoted functions for
            # identification
            setattr(func, '___remoted_by_hoplite___', True)
            setattr(mod_func, '___is_hoplite_remotable___', True)
            setattr(async_mod_func, '___is_hoplite_remotable___', True)
            setattr(module, 'remote_' + name, mod_func)
            setattr(module, 'remote_async_' + name, async_mod_func)
        else:
            raise RuntimeError(
                'Unable to add remote capabilities to object {} which is of'
                ' type {}'.format(obj.__name__, type(obj))
            )
        return obj
    return inner


def _get_remote_docstring(ref_type, namespace, func_name):
    return 'This function calls :{0}:`{1}.{2}` on a remote machine which is ' \
           'running a Hoplite server.\n\n' \
           ':param remote_machine_address: The hostname or IP address of the' \
           'remote machine\n' \
           ':ref_type remote_machine_address: str\n' \
           ':param args: Positional arguments for {2}\n' \
           ':param kwargs: Keyword arguments for {2}\n' \
           ':returns: The value or values returned by {2} after it finishes ' \
           ' running on the remote machine\n\n' \
           'This function raises the same exceptions as {2}. If an error ' \
           'occurs in the Hoplite framework, or if the original exception ' \
           'raised on the remote machine cannot be raised on the local ' \
           'machine, then a JobFailedError (from the Hoplite module) will ' \
           'be raised.'.format(
               ref_type, namespace, func_name)


def _get_remote_async_docstring(ref_type, namespace, func_name):
    return 'This function returns an object which can be used to call ' \
           ':{0}:`{1}.{2}` asynchronously on a remote machine which is ' \
           'running a Hoplite server.  \n\n' \
           ':param remote_machine_address: The hostname or IP address of the' \
           ' remote machine\n' \
           ':ref_type remote_machine_address: str\n' \
           ':param args: Positional arguments for {2}\n' \
           ':param kwargs: Keyword arguments for {2}\n' \
           ':returns: An instance of the RemoteAsyncJobWrapper class, which ' \
           ' can be used to start {2} on the remote machine and query its ' \
           'status. This class implements the same public interface as the ' \
           'RemoteJob class in Hoplite, and therefore provides methods such ' \
           ' as start, join, and finished. When join is called, the class '\
           'will block until the function finishes executing on the remote ' \
           'machine, and then will return the values returned by the ' \
           'remotely executed function.\n' \
           ':rtype: RemoteAsyncJobWrappper\n\n' \
           'This function raises the same exceptions as {2}. If an error ' \
           'occurs in the Hoplite framework, or if the original exception ' \
           'raised on the remote machine cannot be raised on the local ' \
           'machine, then a JobFailedError (from the Hoplite module) will ' \
           'be raised. Note that any exceptions raised on the remote ' \
           'machine will not be called until the status of the job is ' \
           'checked, such as when "finished" or "join" are called on the ' \
           'RemoteAsyncJobWrappper object.'.format(
               ref_type, namespace, func_name)


class RemoteEnablerMetaClass(type):
    """
    .. deprecated:: 15.0.0.dev25
       Use the :ref:`remotify decorator <remotify>` instead

    Add remote capabilities to a class.

    If a class opts to use this as its metaclass, then every function in the
    class (besides those starting with __) will be enhanced to allow for remote
    operation through hoplite. As an example, if there is a function defined
    as::

        def do_stuff(self, input_1, input_2):
            ...

    Then two additional functions will be added to the class::

        def remote_do_stuff(self, remote_machine_address, *args, **kwargs):
            ...

        def remote_async_do_stuff(self, remote_machine_address, *args, **kwargs)
            ...

    In these new functions, \*args and \*\*kwargs represent all of the
    arguments required by the original function. "remote_machine_address" is
    the IP address or hostname of the machine on which the function will be
    run remotely. If the remote machine is running Hoplite on a port other
    than the default (5000), then "remote_machine_address" can be given in the
    form "address:port"

    When the remote_do_stuff function is called, hoplite will attempt to
    connect to the Hoplite server on the remote machine and run the function
    on it, using the current state of the object on the local machine. In
    other words, it should seem as if the function is running on the local
    machine, except that the operations themselves will affect *only* the
    remote machine. In particular, if the function being run remotely changes
    the class instance, those changes will not be reflected on the local
    machine. This must be kept in mind when creating classes that will be made
    remotable - any changes of state must be sent back to the local machine as
    return values. That includes reference variables passed as function
    parameters that would normally not need to be returned.

    When the remote_async_do_stuff function is called, a RemoteAsyncJobWrapper
    object will be returned which can be used to start the function
    asynchronously. This object implements the same public functions as the
    RemoteJob class, and so can be used in the same way. When the function is
    run, it operates in the same way as the remote_do_stuff function does,
    except that it is run asynchronously. This means that any exceptions which
    occur will not be raised until the job status is checked.

    In cases where inheritance is used in a class, the behavior is governed by
    the following rules:

    - For each class which uses the metaclass, if it inherits from another
    class then that parent class must either use the metaclass as well (in
    which case all of its functions will be made available to the child as
    remotable functions) or the parent class must be a new-style class
    (inherit from "object"), in which case its functions will not be
    available to the child as remotable functions. If the parent class does
    not use the metaclass, and if it is not a new-style class, then a TypeError
    will be raised due to some technical issues with metaclasses and
    inheritance.

    - If a class uses the metaclass, then all classes which inherit from it
    (i.e. all of its descendants) will be made remotable. In other words, if a
    parent class uses the metaclass, then it as well as its child will be made
    remotable. The parent will have access to all its own functions as
    remotable functions, and the child will have access not only to all of its
    own functions, but also all of the parent's functions, as remotable
    functions. This holds true for all descendants.

    :returns: The class, enhanced to allow for remote functionality
    """
    def __new__(mcs, clsname, bases, dct):
        for name, val in list(dct.items()):
            if (not name.startswith('__')) and hasattr(val, '__call__'):
                # Function must be built in a separate function and then
                # assigned here.  Otherwise, there are problems with each
                # function added to the class actually pointing to the same
                # thing.
                if name.startswith('remote_') or name.startswith('async_'):
                    raise AttributeError(
                        'Unable to add remote capability to function {0}:'
                        ' function cannot begin with "remote_" or '
                        '"async_"'.format(name))
                dct['remote_' + name] = remote_func_builder(name)
                dct['remote_async_' + name] = remote_async_func_builder(name)
        return type.__new__(mcs, clsname, bases, dct)


def remote_func_builder(function_name):
    """
    Build a function that will connect to a remote machine and execute a
    function on it.

    :param function_name: The name of the class function that will be called on
        the remote machine.  This is necessary because, even though it is
        technically something like remote_do_stuff that is called, it will be
        recognized as _remote_func instead.
    :returns: Function that, when called, will connect to a remote machine and
        execute the function represented by 'function_name'
    """

    def _remote_func(self, remote_machine_address, *args, **kwargs):
        """
        Call a function on a remote machine.

        Note that the class instance is pickled and sent to the remote machine.
        This is so the current state of the class will be utilized when the
        function is called on the remote machine. There are probably lots of
        corner cases in which this will cause problems. For example, you must
        be aware of side effects of the original function that you might be
        expecting to affect the local machine, because they will not do so.

        :param remote_machine_address: IP address or hostname of the remote
            machine on which the function will be run.
            If the remote machine is running Hoplite on a port other than the
            default (5000), then "remote_machine_address" can be specified in
            the form "address:port"
        :param remote_timeout: Timeout (in floating-point seconds) of the
            function
        :param *args: Normal arguments being passed to the remotely called
            function
        :param *kwargs: Keyword arguments being passed to the remotely called
            function
        :returns: The value(s) returned by the function which was called on the
            remote machine
        """
        logger = logging.getLogger(__name__)
        logger.addHandler(logging.NullHandler())

        remote_timeout = -1
        timeout_message = ''
        if kwargs.get('remote_timeout') is not None and kwargs['remote_timeout'] > 0.0:
            remote_timeout = kwargs['remote_timeout']
            kwargs.pop('remote_timeout')
            timeout_message = ' with timeout of {} seconds'.format(
                remote_timeout)

        args_string = args.__str__()
        kwargs_string = kwargs.__str__()
        if not HopliteClientSettings.debug:
            if len(args_string) > 33:
                args_string = args_string[0:30] + '...'
            if len(kwargs_string) > 33:
                kwargs_string = kwargs_string[0:30] + '...'
        logger.info('"{0}" on target "{1}" with args: {2} and '
                    'kwargs: {3}{4}'.format(
                        function_name,
                        remote_machine_address,
                        args_string,
                        kwargs_string,
                        timeout_message))
        config = {
            'args': pickle.dumps(args),
            'kwargs': pickle.dumps(kwargs),
            'instance': pickle.dumps(self),
            'function_name': function_name
        }
        job = None
        try:
            job_manager = client.remote_job_manager.RemoteJobManager(
                remote_machine_address)
            job = job_manager.create_job(
                'hoplite.plugins.remote_enabler_job', config)
            job.start()
            job.join(remote_timeout)

        except JobFailedError as e:
            if job is None:
                logger.error(
                    'Exception occurred while creating job to call "{0}" on'
                    ' "{1}": {2}'.format(
                        function_name, remote_machine_address, str(e))
                )
            else:
                logger.error(
                  'Exception occurred while calling "{0}" on '
                  ' "{1}": {2}'.format(
                      function_name,
                      remote_machine_address,
                      e.__str__())
                )
            # ALL TRACEBACK ENTRIES BELOW THIS ARE FROM THE REMOTE MACHINE
            e.raise_remote_exception()

        return_values = pickle.loads(job.status().get('return_values'))

        if return_values is None:
            return None

        # Convert return value into either a single value, or a tuple, so that
        # it appears the same as if the function were called on the local
        # machine
        if len(return_values) > 1:
            return_object = tuple(return_values)
        else:
            return_object = return_values[0]

        return_object_string = return_object.__str__()
        # limit return object string if not debugging
        if not HopliteClientSettings.debug and len(return_object_string) > 50:
            return_object_string = return_object_string[0:47] + '...'

        logger.debug('"{0}" on target "{1}" returned {2}'.format(
            function_name,
            remote_machine_address,
            return_object_string))
        return return_object
    return _remote_func


def remote_async_func_builder(function_name):
    """
    Build a function that will connect to a remote machine and create a job
    wrapper that can be used to run a function asynchronously

    :param function_name: The name of the class function that will be called on
        the remote machine.
    :returns: Function that, when called, will connect to a remote machine and
        create then return a job wrapper for running the specified function
    """

    def _remote_async_func(self, remote_machine_address, *args, **kwargs):
        """
        Create a job on a remote machine and return a wrapper so that it has
        the same interface as a job running on the local machine.

        ***NOTE*** It seems that join must always be eventually called on jobs,
        even if they have already finished. Otherwise, a process gets orphaned,
        causing problems with future operations.

        :param remote_machine_address: IP address or hostname of the remote
            machine on which the function will be run.  If the remote machine
            is running Hoplite on a port other than the default (5000), then
            the port can be specified in the form "address:port"
        :param *args: Normal arguments being passed to the remotely called
            function
        :param *kwargs: Keyword arguments being passed to the remotely called
            function
        :returns: A wrapper around the remote job
        """
        logger = logging.getLogger(__name__)
        logger.addHandler(logging.NullHandler())

        logger.debug(
            'Creating job "{0}" on target "{1}" with args: {2} and kwargs:'
            ' {3}'.format(
                function_name, remote_machine_address, args, kwargs)
        )
        config = {
            'args': pickle.dumps(args),
            'kwargs': pickle.dumps(kwargs),
            'instance': pickle.dumps(self),
            'function_name': function_name
        }
        job_manager = client.remote_job_manager.RemoteJobManager(
            remote_machine_address)
        job = job_manager.create_job(
            'hoplite.plugins.remote_enabler_job', config)
        return RemoteAsyncJobWrapper(job, function_name)
    return _remote_async_func


def remote_module_func_builder(function_name, module_name):
    """
    Build a function that will connect to a remote machine and execute a
    function on it.

    :param function_name: The name of the function that will be called on the
        remote machine.
    :returns: Function that, when called, will connect to a remote machine and
        execute the function represented by 'function_name'
    """
    def _remote_module_func(remote_machine_address, *args, **kwargs):
        """
        Call a function on a remote machine.

        Note that the class instance is pickled and sent to the remote machine.
        This is so the current state of the class will be utilized when the
        function is called on the remote machine. There are probably lots of
        corner cases in which this will cause problems, so be aware of side
        effects of the original function.

        :param remote_machine_address: IP address or hostname of the remote
            machine on which the function will be run.  If the remote machine
            is running Hoplite on a port other than the default (5000), then
            the port can be specified in the form "address:port"
        :param *args: Normal arguments being passed to the remotely called
            function
        :param remote_timeout:
        :param *kwargs: Keyword arguments being passed to the remotely called
            function
        :returns: The value(s) returned by the function which was called on the
            remote machine
        """
        logger = logging.getLogger(__name__)
        logger.addHandler(logging.NullHandler())

        remote_timeout = -1
        timeout_message = ''
        if kwargs.get('remote_timeout') is not None and kwargs['remote_timeout'] > 0.0:
            remote_timeout = kwargs['remote_timeout']
            kwargs.pop('remote_timeout')
            timeout_message = ' with timeout of {} seconds'.format(
                remote_timeout)

        args_string = args.__str__()
        kwargs_string = kwargs.__str__()
        if not HopliteClientSettings.debug:
            if len(args_string) > 33:
                args_string = args_string[0:30] + '...'
            if len(kwargs_string) > 33:
                kwargs_string = kwargs_string[0:30] + '...'
        logger.info('"{0}" on target "{1}" with args: {2} and '
                    'kwargs: {3} {4}'.format(
                        function_name,
                        remote_machine_address,
                        args_string,
                        kwargs_string,
                        timeout_message))
        config = {
            'args': pickle.dumps(args),
            'kwargs': pickle.dumps(kwargs),
            'module_name': module_name,
            'function_name': function_name
        }
        job = None
        try:
            job_manager = client.remote_job_manager.RemoteJobManager(
                remote_machine_address)
            job = job_manager.create_job(
                'hoplite.plugins.remote_enabler_module_job', config)
            job.start()
            job.join(remote_timeout)
        except JobFailedError as e:
            if job is None:
                logger.error(
                    'Exception occurred while creating job to call "{0}" '
                    'on "{1}": {2}'.format(
                        function_name, remote_machine_address, str(e))
                )
            else:
                logger.error(
                    'Exception occurred while calling "{0}" on "{1}": '
                    '{2}'.format(
                        function_name, remote_machine_address, e.__str__())
                )
            # ALL TRACEBACK ENTRIES BELOW THIS ARE FROM THE REMOTE MACHINE
            e.raise_remote_exception()

        return_values = pickle.loads(job.status().get('return_values'))

        if return_values is None:
            return None

        # Convert return value into either a single value, or a tuple, so that
        # it appears the same as if the function were called on the local
        # machine
        if len(return_values) > 1:
            return_object = tuple(return_values)
        else:
            return_object = return_values[0]

        return_object_string = return_object.__str__()
        # limit return object string if not debugging
        if not HopliteClientSettings.debug and len(return_object_string) > 50:
            return_object_string = return_object_string[0:47] + '...'

        logger.debug(
            '"{0}" on target "{1}" returned {2}'.format(function_name,
                                                        remote_machine_address,
                                                        return_object_string))
        return return_object
    return _remote_module_func


def remote_module_async_func_builder(function_name, module_name):
    """
    Build a function that will connect to a remote machine and create a job
    wrapper that can be used to run a function asynchronously

    :param function_name: The name of the class function that will be called on
        the remote machine.
    :returns: Function that, when called, will connect to a remote machine and
        create then return a job wrapper for running the specified function
    """
    def _remote_async_module_func(remote_machine_address, *args, **kwargs):
        """
        Create a job on a remote machine and return a wrapper so that it has
        the same interface as a job running on the local machine.

        ***NOTE*** It seems that join must always be eventually called on jobs,
        even if they have already finished. Otherwise, a process gets orphaned,
        causing problems with future operations.

        :param remote_machine_address: IP address or hostname of the remote
            machine on which the function will be run. If the remote machine is
            running Hoplite on a port other than the default (5000), then the
            port can be specified in the form "address:port"
        :param *args: Normal arguments being passed to the remotely called
            function
        :param *kwargs: Keyword arguments being passed to the remotely called
            function
        :returns: A wrapper around the remote job
        """
        logger = logging.getLogger(__name__)
        logger.addHandler(logging.NullHandler())
        args_string = args.__str__()
        kwargs_string = kwargs.__str__()
        if not HopliteClientSettings.debug:
            if len(args_string) > 33:
                args_string = args_string[0:30] + '...'
            if len(kwargs_string) > 33:
                kwargs_string = kwargs_string[0:30] + '...'
        logger.info('Creating job "{0}" on target "{1}" with args: {2} and '
                    'kwargs: {3}'.format(
                        function_name,
                        remote_machine_address,
                        args_string,
                        kwargs_string))
        config = {
            'args': pickle.dumps(args),
            'kwargs': pickle.dumps(kwargs),
            'module_name': module_name,
            'function_name': function_name
        }
        job_manager = client.remote_job_manager.RemoteJobManager(
            remote_machine_address)
        job = job_manager.create_job(
            'hoplite.plugins.remote_enabler_module_job', config)
        return RemoteAsyncJobWrapper(job, function_name)
    return _remote_async_module_func


class RemoteAsyncJobWrapper:
    """
    This class is a wrapper around the RemoteJob class, and is used for
    asynchronously running functions which are called remotely on another
    machine. It implements the same public methods as the RemoteJob class,
    and the reader should refer to that module for additional information on
    how to use it.
    """

    def __init__(self, job, function_name):
        self.job = job
        self.function_name = function_name
        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(logging.NullHandler())

    def start(self):
        """
        Start running the function.
        """

        self.logger.debug(
            'Starting "{0}({1})" on "{2}:{3}"'.format(self.function_name,
                                                      self.job.uuid,
                                                      self.job.address,
                                                      self.job.port))
        self.job.start()

    def join(self, remote_timeout=-1):
        """
        Join the function once it has been started.

        This differs from the join function in RemoteJob in that, upon
        completion, it returns the values returned by the remotely executed
        function.

        :param remote_timeout: Optional timeout in seconds. -1 Implies no
            timeout
        :type remote_timeout: int
        :return: Values returned by the remotely executed function
        """
        self.logger.debug('Joining "{0}" on "{1}:{2}"'.format(
            self.function_name, self.job.address, self.job.port))
        try:
            self.job.join(remote_timeout)
        except JobFailedError as e:
            self.logger.error(
                'Exception occurred while calling "{0}" on target "{1}":'
                ' {2}'.format(
                    self.function_name, self.job.address, e.__str__())
            )
            # ALL TRACEBACK ENTRIES BELOW THIS ARE FROM THE REMOTE MACHINE
            e.raise_remote_exception()
        return_values = pickle.loads(self.job.status().get('return_values'))

        if return_values is None:
            return None

        # Convert return value into either a single value, or a tuple, so that
        # it appears the same as if the function were called on the local
        # machine
        if len(return_values) > 1:
            return_object = tuple(return_values)
        else:
            return_object = return_values[0]

        return_object_string = return_object.__str__()
        if not HopliteClientSettings.debug and (len(return_object_string) > 50):
            return_object_string = return_object_string[0:47] + '...'

        self.logger.debug(
            '"{0}" on target "{1}:{2}" returned {3}'.format(
                self.function_name,
                self.job.address,
                self.job.port,
                return_object_string))
        return return_object

    def config(self, force=False):
        """
        Get the configuration dictionary for the job.
        """
        return self.job.config(force)

    def status(self, force=False):
        """
        Check the status of the function's execution. This will raise an
        exception if the function has encountered an error since the last time
        the status was checked.
        :param force:
        :return:
        """
        return self.job.status(force)

    def kill(self, force=False):
        """
        Attempt to terminate the job.
        """
        return self.job.kill(force)

    def running(self, force=False):
        """
        Check if the job is still running.
        """
        return self.job.running(force)

    def finished(self, force=False):
        """
        Check if the job is finished.
        """
        return self.job.finished(force)


# This is used as an example of how remoted functions get automatically
# documented. Do not use.
@remotify(__name__)
def my_func(arg1, arg2):
    """Function which does stuff.

    My_func is a function which does nothing that is particularly useful. It
    serves primarily as an example.

    :param arg1: A number which will be printed to the console
    :type arg1: int
    :param arg2: A string which will also be printed
    :type arg2: str
    :returns: The concatenated number and string
    """
    concat_string = str(arg1) + arg2
    print(concat_string)
    return concat_string


# This is used as an example of how remoted classes get automatically
# documented. Do not use.
@remotify(__name__)
class Foo(object):
    def __init__(self, val_1):
        """Initialize a new instance of the Foo class.

        :param val_1: Any value. Doesn't matter what value it is
        """
        self.val_1 = val_1

    def print_a_val(self):
        """Prints a value.

        Prints the value that was passed in when the class instance was
        initialized.
        """
        print((self.val_1))

    def print_another_val(self, another_val):
        """Prints another value.

        Prints a value passed in by the user. It can be the same as the other
        value if desired.

        :param another_val: Another value. Doesn't matter what value it is
        """
        print(another_val)



from hoplite.client.helpers import ClientMixin
import unittest2
from hoplite.serializer import hoplite_dumps
from httmock import HTTMock, urlmatch, response, all_requests
from tests.utils import StatusCodeTestMixin
from hoplite.exceptions import InternalServerError

DATA = {"status": "ok"}

def _request_has_json_content_type(request):
    if request.headers['content-type'] == 'application/json':
        return True
    return False

@urlmatch(netloc='localhost:5001', path='\/test$')
def get(url, request):
    if request.method == 'GET':
        return response(200)

@urlmatch(netloc='localhost:5001', path='\/test$')
def post(url, request):
    print request.body
    if request.method == 'POST' and _request_has_json_content_type(request)\
            and request.body == hoplite_dumps(DATA):
        return response(200)

@urlmatch(netloc='localhost:5001', path='\/test$')
def put(url, request):
    if request.method == 'PUT' and _request_has_json_content_type(request)\
            and request.body == hoplite_dumps(DATA):
        return response(200)

@urlmatch(netloc='localhost:5001', path='\/test$')
def patch(url, request):
    if request.method == 'PATCH' and _request_has_json_content_type(request)\
            and request.body == hoplite_dumps(DATA):
        return response(200)

@urlmatch(netloc='localhost:5001', path='\/test$')
def delete(url, request):
    if request.method == 'DELETE' and _request_has_json_content_type(request)\
            and request.body == hoplite_dumps(DATA):
        return response(200)

def return_500(url, request):
    response(500)

class HopliteTestCase(unittest2.TestCase):
    pass

class HopliteClientTestCase(StatusCodeTestMixin, HopliteTestCase):
    pass

class TestClientMixin(HopliteClientTestCase):
    def setUp(self):
        self.mixin = ClientMixin()

    def test_jget(self):
        with HTTMock(get):
            r = self.mixin.jget("http://localhost:5001/test")
            self.assertOk(r)

    def test_jpost(self):
        with HTTMock(post):
            r = self.mixin.jpost("http://localhost:5001/test", data=DATA)
            self.assertOk(r)

    def test_jput(self):
        with HTTMock(put):
            r = self.mixin.jput("http://localhost:5001/test", data=DATA)
            self.assertOk(r)

    def test_jpatch(self):
        with HTTMock(patch):
            r = self.mixin.jpatch("http://localhost:5001/test", data=DATA)
            self.assertOk(r)

    def test_jdelete(self):
        with HTTMock(delete):
            r = self.mixin.jdelete("http://localhost:5001/test", data=DATA)
            self.assertOk(r)

    def raises_on_500_status_code(self):
        with HTTMock(return_500):
            self.assertRaises(InternalServerError, self.mixin.jget("localhost"))
            self.assertRaises(InternalServerError, self.mixin.jpost("localhost"))
            self.assertRaises(InternalServerError, self.mixin.jput("localhost"))
            self.assertRaises(InternalServerError, self.mixin.jpatch("localhost"))
            self.assertRaises(InternalServerError, self.mixin.jdelete("localhost"))


"""
Modules for classes that are used as containers for global variables.
Any function can import and modify/read the values stored in these classes.
"""


class HopliteServerSettings():
    """
    Holds global hoplite settings for the server.
    """
    # Debug mode increases the amount of logging performed by certain
    # functions.
    debug = False


class HopliteClientSettings():
    """
    Holds global hoplite settings for the client.
    """
    # Debug mode increases the amount of logging performed by certain
    # functions.
    debug = False



from hoplite.utils import server_logging
from flask import json
from hoplite.api.helpers import job_manager
from . import HopliteApiTestCase

logger = server_logging.get_server_logger(__name__)


class JobPluginsApiTestCase(HopliteApiTestCase):
    def setUp(self):
        super(JobPluginsApiTestCase, self).setUp()
        self.manager = job_manager

    def test_get_job_plugins(self):
        r = self.client.get('/job_plugins')
        # 200 OK
        self.assertOk(r)
        # Test that all test builtin_plugins are returned
        body = json.loads(r.get_data())
        job_plugins = body["job_plugins"]
        self.assertEquals(len(job_plugins), 4)
        expected = [self.test_jobs_module.constants.THROW_AN_EXCEPTION_JOB_NAME,
                    self.test_jobs_module.constants.CREATE_FILE_JOB_NAME,
                    self.test_jobs_module.constants.WAIT_10_SECONDS_JOB_NAME,
                    self.test_jobs_module.constants.JOB_FAILED_EXCEPTION_JOB_NAME]

        self.assertEquals(sorted(job_plugins), sorted(expected))


from hoplite.utils import server_logging

logger = server_logging.get_server_logger(__name__)
from hoplite.server.jobs.job_manager import JobManager
from hoplite.plugin_manager import EntryPointManager
from hoplite.exceptions import JobPluginDoesNotExistError, JobDoesNotExistError
from tests import HopliteTestCase


class TestJobManager(HopliteTestCase):
    def setUp(self):
        super(TestJobManager, self).setUp()
        self.manager = JobManager(EntryPointManager('hoplite.test_jobs'))

    def test_available_jobs(self):
        job_list = self.manager.available_job_plugins()
        self.assertEquals(len(job_list), 4)
        expected = [self.test_jobs_module.constants.THROW_AN_EXCEPTION_JOB_NAME,
                    self.test_jobs_module.constants.CREATE_FILE_JOB_NAME,
                    self.test_jobs_module.constants.WAIT_10_SECONDS_JOB_NAME,
                    self.test_jobs_module.constants.JOB_FAILED_EXCEPTION_JOB_NAME]
        self.assertEquals(sorted(job_list), sorted(expected))

    def test_job_info_raises_on_invalid_id(self):
        # Rases an exception when passed a bad job id
        self.assertRaises(JobDoesNotExistError, self.manager.get_job, 1)

    def test_job_info_gets_correct_job(self):
        job = self.manager.create_job(self.test_jobs_module.constants.WAIT_10_SECONDS_JOB_NAME, {}, port=5001)
        job_2 = self.manager.get_job(job.uuid)
        self.assertEquals(job, job_2)

    def test_create_job_raises_on_invalid_job_plugin_name(self):
        with self.assertRaises(JobPluginDoesNotExistError):
            self.manager.create_job("Fake Job Name", {}, port=5001)

    def test_create_job_running_false(self):
        job = self.manager.create_job(self.test_jobs_module.constants.THROW_AN_EXCEPTION_JOB_NAME, {}, port=5001)
        self.assertIsNotNone(job.uuid)
        self.assertFalse(job.running())

    def test_create_job_running_true_runs_job(self):
        job = self.manager.create_job(self.test_jobs_module.constants.WAIT_10_SECONDS_JOB_NAME, {}, True, port=5001)
        self.assertIsNotNone(job.uuid)
        self.assertTrue(job.running())
        job.kill()


from hoplite.utils import server_logging
from tests.api import HopliteApiTestCase
from flask import json
from hoplite.api.helpers import job_manager
import time

logger = server_logging.get_server_logger(__name__)


class JobsApiTestCase(HopliteApiTestCase):
    def setUp(self):
        super(JobsApiTestCase, self).setUp()
        self.manager = job_manager

    def test_get_jobs(self):
        self.manager.create_job(self.test_jobs_module.constants.WAIT_10_SECONDS_JOB_NAME, {})
        self.manager.create_job(self.test_jobs_module.constants.THROW_AN_EXCEPTION_JOB_NAME, {})
        r = self.client.get('/jobs')
        self.assertOk(r)
        job_list = json.loads(r.get_data())["jobs"]
        print job_list
        self.assertEquals(len(job_list), 2)
        # Get names from 3 jobs then sort and compare
        names = []
        for job in job_list:
            names.append(job["name"])
        expected = [self.test_jobs_module.constants.WAIT_10_SECONDS_JOB_NAME,
                    self.test_jobs_module.constants.THROW_AN_EXCEPTION_JOB_NAME]
        self.assertEquals(sorted(names), sorted(expected))

    def test_post_jobs_blank_config(self):
        data = {"name": self.test_jobs_module.constants.WAIT_10_SECONDS_JOB_NAME}
        r = self.jpost('/jobs', data=data)
        self.assertOk(r)
        r_job = json.loads(r.get_data())
        self.assertEquals(data["name"], r_job["name"])
        self.assertEquals(r_job["config"], {})
        self.assertEquals(r_job["running"], False)

    def test_post_jobs_config(self):
        data = {"name": self.test_jobs_module.constants.WAIT_10_SECONDS_JOB_NAME, "config": { "something": "yay" }}
        r = self.jpost('/jobs', data=data)
        self.assertOk(r)
        r_job = json.loads(r.get_data())
        self.assertEquals(data["name"], r_job["name"])
        self.assertEquals(data["config"], r_job["config"])
        self.assertEquals(r_job["running"], False)

    def test_post_jobs_running_true(self):
        data = {"name": self.test_jobs_module.constants.WAIT_10_SECONDS_JOB_NAME, "running": True}
        r = self.jpost('/jobs', data=data)
        self.assertOk(r)
        r_job = json.loads(r.get_data())
        self.assertEquals(data["name"], r_job["name"])
        self.assertEquals(r_job["config"], {})
        self.assertEquals(r_job["running"], True)
        self._terminate_all_jobs()

    def test_post_job_bad_name_returns_400(self):
        data = {"name": "Bad Name", "config": { "something": "yay" }}
        r = self.jpost('/jobs', data=data)
        self.assertBadRequest(r)
        self.assertEquals(json.loads(r.get_data())["error"], "Job plugin 'Bad Name' does not exist")

    def test_put_job_updates_status(self):
        job = self._create_job()
        status_update = { "status": { "my_status": "is good" }, "api_key": job._api_key}
        r = self.jput("/jobs/{0}".format(job.uuid), data=status_update)
        self.assertEquals(job.status(), { "my_status": "is good" })

    def test_get_jobs_running(self):
        e_job_uuid = []
        jobs = []
        print self.manager.available_job_plugins()
        for i in range(4):
            job = self.manager.create_job(self.test_jobs_module.constants.WAIT_10_SECONDS_JOB_NAME, {}, True)
            jobs.append(job)
            e_job_uuid.append(job.uuid)
        print jobs[0].status()
        r = self.client.get('/jobs/running')
        self.assertOk(r)
        r_jobs = json.loads(r.get_data())["jobs"]
        self.assertEquals(len(r_jobs), 4)
        r_job_uuid = []
        for job in r_jobs:
            r_job_uuid.append(job["uuid"])
        self.assertEquals(sorted(e_job_uuid), sorted(r_job_uuid))
        self._terminate_all_jobs()

    def test_put_start_job(self):
        job = self.manager.create_job(self.test_jobs_module.constants.WAIT_10_SECONDS_JOB_NAME, {})
        self.assertFalse(job.running())
        r = self.client.put('/jobs/{0}/start'.format(job.uuid))
        self.assertOk(r)
        r_body = json.loads(r.get_data())
        self.assertTrue(r_body["started"])
        self.assertEquals(r_body["uuid"], job.uuid)
        self.assertTrue(job.running())
        self._terminate_all_jobs()

    def test_put_start_job_error_if_job_finished(self):
        job = self._create_job(self.test_jobs_module.constants.THROW_AN_EXCEPTION_JOB_NAME, running=True)
        while job.running():
            pass
        r = self.client.put('/jobs/{0}/start'.format(job.uuid))
        self.assertForbidden(r)
        r_body = json.loads(r.get_data())
        self.assertEquals(r_body["error"], "Job UUID: {0} you cannot start a job more than once".format(job.uuid))

    def test_put_start_job_job_already_started(self):
        job = self._create_job(running=True)
        self.assertTrue(job.running())
        r = self.client.put('/jobs/{0}/start'.format(job.uuid))
        self.assertForbidden(r)
        r_body = json.loads(r.get_data())
        self.assertEquals(r_body["error"], "Job UUID: {0} you cannot start a job more than once".format(job.uuid))
        self._terminate_all_jobs()

    def test_put_start_job_error_if_job_does_not_exist(self):
        r = self.client.put('/jobs/{0}/start'.format(3288283))
        self.assertNotFound(r)
        r_body = json.loads(r.get_data())
        self.assertEquals(r_body["error"], "Job with UUID: 3288283 does not exist")

    def test_put_kill_job_error_if_job_finished(self):
        job = self._create_job(self.test_jobs_module.constants.THROW_AN_EXCEPTION_JOB_NAME, running=True)
        while job.running():
            pass
        r = self.client.put('/jobs/{0}/kill'.format(job.uuid))
        self.assertOk(r)
        r_body = json.loads(r.get_data())

    def test_put_kill_job_error_if_job_not_started(self):
        job = self._create_job()
        r = self.client.put('/jobs/{0}/kill'.format(job.uuid))
        self.assertForbidden(r)
        r_body = json.loads(r.get_data())
        self.assertEquals(r_body["error"], "Job UUID: {0} has not been started".format(job.uuid))

    def test_put_kill_job_error_if_job_does_not_exist(self):
        r = self.client.put('/jobs/{0}/kill'.format(12312))
        self.assertNotFound(r)
        r_body = json.loads(r.get_data())
        self.assertEquals(r_body["error"], "Job with UUID: 12312 does not exist")

    def test_put_kill_job(self):
        job = self.manager.create_job(self.test_jobs_module.constants.WAIT_10_SECONDS_JOB_NAME, {}, True)
        self.assertTrue(job.running())
        r = self.client.put('/jobs/{0}/kill'.format(job.uuid))
        self.assertOk(r)
        r_body = json.loads(r.get_data())
        self.assertTrue(r_body["killed"])
        self.assertEquals(r_body["uuid"], job.uuid)
        start_time = time.time()
        while job.running():
            elapsed_time = time.time() - start_time
            if elapsed_time > 5:
                self.fail("Timed out waiting for job to die")
        self.assertTrue(job.finished())

    def test_get_jobs_uuid(self):
        job = self._create_job()
        r = self.client.get('/jobs/{0}'.format(job.uuid))
        self.assertOk(r)
        r_job = json.loads(r.get_data())
        self.assertEquals(r_job["uuid"], job.uuid)
        self.assertEquals(r_job["name"], self.test_jobs_module.constants.WAIT_10_SECONDS_JOB_NAME)
        self.assertFalse(r_job["running"])
        self.assertFalse(r_job["finished"])
        self.assertFalse(r_job["killed"])
        logger.info("Starting job")
        job.start()
        r = self.client.get('/jobs/{0}'.format(job.uuid))
        self.assertOk(r)
        r_job = json.loads(r.get_data())
        self.assertTrue(r_job["running"])
        self.assertFalse(r_job["finished"])
        self.assertFalse(r_job["killed"])
        self._terminate_all_jobs()
        while job.running():
            pass
        r = self.client.get('/jobs/{0}'.format(job.uuid))
        self.assertOk(r)
        r_job = json.loads(r.get_data())
        self.assertFalse(r_job["running"])
        self.assertTrue(r_job["finished"])
        self.assertTrue(r_job["killed"])

    def test_get_jobs_uuid_with_bad_id(self):
        r = self.client.get('/jobs/{0}'.format("8b7fea59-2c0d-4afa-8109-2bc0a26ec865"))
        self.assertNotFound(r)
        self.assertEquals(json.loads(r.get_data())["error"], "Job with UUID: 8b7fea59-2c0d-4afa-8109-2bc0a26ec865 does not exist")

    def _create_job(self, name=None, config={}, running=False):
        if name is None:
            name = self.test_jobs_module.constants.WAIT_10_SECONDS_JOB_NAME
        return self.manager.create_job(name, config, running, port=5001)

    def _terminate_all_jobs(self):
        jobs = self.manager.all_jobs()
        for job in jobs:
            job.kill()



import os

TESTS_PATH = os.path.split(__file__)[0]
TEST_RESOURCES_PATH = os.path.join(TESTS_PATH, "test_resources")
FTP_TEST_DIR_PATH = os.path.join(TEST_RESOURCES_PATH, "ftp_test")
TEST_PACKAGE_PATH = os.path.join(TEST_RESOURCES_PATH, "test_jobs_package")
TEST_ENTRY_POINT_PATH = os.path.join(TEST_RESOURCES_PATH, "test_entry_point")


import werkzeug
from werkzeug.local import LocalProxy
from hoplite.serializer import hoplite_dumps


# Jsonify that uses hoplite's serial decoder
def jsonify(*args, **kwargs):
    """
        jsonify that uses hoplite's serial encoder
    """
    return werkzeug.Response(
        hoplite_dumps(dict(*args, **kwargs)), mimetype='application/json')


# This gets set by the app factory when the app is created
manager = None


def get_job_manager():
    return manager

job_manager = LocalProxy(get_job_manager)



"""
Overview
========

Hoplite's main purpose is to make it easy to remotely invoke python code.
It does this by loading all modules registered with the setuptool's entry point
'hoplite.jobs'. These 'jobs' satisfy a simple interface, which allows Hoplite
to run them, and to report on their status. For more information on building
modules to plug into Hoplite see :ref:`job-plugins`.

Since creating plugins for simple tasks is cumbersome, Hoplite also provides
a very simple way of running existing Python code on remote machines. This
functionality is provided in the form of a decorator which, when applied to a
function or class, adds the ability to execute the function or class methods
on the remote machine. Detailed information on how to do this is provided
:ref:`here <remote-enabler>`.

Hoplite is split into a server and a client. The server can be invoked via the
command line with :ref:`hoplite-server` and the client is invoked with
:ref:`hoplite-client`. The client is used to manage local and
remote hoplite-server instances.
"""
from hoplite.remote_enabler import remotify
from hoplite.public_api import (
    is_hoplite_available,
    wait_for_hoplite,
    remote_install_python_package,
    remote_uninstall_python_package)



"""
Role
====
Used to communicate with a remote hoplite server. This can be used to create
jobs and to manage the job plugin search paths. This is a wrapper over the
:ref:`REST-API-Server` and :ref:`REST-API-Job-Plugins`.

API
===
"""
import logging
import socket
import time
from hoplite.client.helpers import ClientMixin
from hoplite.client.remote_job import RemoteJob
from hoplite.serializer import hoplite_loads
from hoplite.exceptions import TimeoutError

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class RemoteJobManager(ClientMixin):
    """
    Used to communicate with a remote hoplite server
    """
    def __init__(self, address, port=5000):
        """
        :param address: IP address or hostname of the remote computer. If
            desired, the address may be in the form "address:port", rather than
            specifying the port in the second parameter
        :param port: The port the hoplite server is listening on. This is
            ignored if the address includes the port
        """
        if ':' in address:
            self.address = address.split(':')[0]
            self.port = address.split(':')[1]
        else:
            self.address = address
            self.port = port
        self._daemon_addr = 'http://{0}:{1}'.format(self.address, self.port)

    def is_available(self):
        """
        Find if the hoplite process is accessible on the remote target

        :return: True if hoplite is available and listening on the target
        :rtype: Boolean
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.connect((self.address, self.port))
            s.close()
        except socket.error:
            return False
        return True

    def wait_for_available(self, retry_period_s=2, retries=150):
        """
        Waits for Hoplite to be active. If it finds hoplite is active before
        the timeout.

        :param int retry_period_s: The number of seconds between retying to
            connect to Hoplite
        :param int retries: The number of attempts to connect to Hoplite to
            make before throwing an exception
        """
        estimated_units = "seconds"
        estimated_timeout = retry_period_s * retries
        if estimated_timeout >= 60:
            estimated_timeout = estimated_timeout/60
            estimated_units = "minutes"

        while True:
            if self.is_available():
                break
            time.sleep(retry_period_s)
            retries -= 1
            if retries <= 0:
                errmsg = 'Hoplite Server could not be reached on host "{0}"'\
                         ' after waiting for at least {1} {2}'.format(
                                self._daemon_addr,
                                estimated_timeout,
                                estimated_units)
                logger.error(errmsg)
                raise TimeoutError(errmsg)

    def get_job(self, uuid):
        """
        Get the job identified by the uuid

        :param str uuid: UUID of the job to get
        :return: the job that matches the uuid
        :rtype: :py:class:`hoplite.client.RemoteJob`
        """
        return RemoteJob(self.address, self.port, uuid=uuid)

    def create_job(self, plugin_name, config):
        """
        Create a job

        :param str plugin_name: name of the plugin you want to run in the job
        :param dict config: the configuration dictionary for the job
        :return: a RemoteJob to access the created job with
        :rtype: :py:class:`hoplite.client.RemoteJob`
        """
        return RemoteJob(
            self.address, self.port, name=plugin_name, config=config)

    def get_running_jobs(self):
        """
        Get a list of jobs that are currently running
        """
        remote = self._daemon_addr + '/jobs/running'
        r = self.jget(remote)
        return hoplite_loads(str(r.text))["jobs"]

    def get_job_plugins(self):
        """
        Get the list of job builtin_plugins that have been loaded

        :return: list of available job_plugins
        :rtype: list of strings
        """
        remote = self._daemon_addr + '/job_plugins'
        r = self.jget(remote)
        return hoplite_loads(str(r.text))["job_plugins"]

    def reload_site_packages(self):
        """
        Force a reload of the local site-packages directory
        that the server uses to look for installed packages

        This allows you to have hoplite load entry points from
        newly installed packages without restarting the hoplite
        instance
        """
        remote = self._daemon_addr + '/reload'
        _ = self.jput(remote)
        return True



"""
Sync down a path from perforce.

Example config::

    {
        KEYS.USER: "username",
        KEYS.P4_PORT: "perforce:1666",
        KEYS.PERFORCE_PATH: "//depot/path",
        KEYS.CLIENT_WORKSPACE: "client_spec_name",
        KEYS.FORCE: True
    }

Example return status::

    No Error:
    {
        "succeeded": True,
        "local_path": "C:\\path",  # The local path the perforce path was synced to
        "stdout": "All the stdout output" # Everything from stdout
    }

    Error:

    {
        "succeeded": False,
        "stdout": "All stdout",
        "stderr": "All stderr"
    }

"""
from hoplite.builtin_plugins.constants import (
    SyncPerforcePathJobConstants as KEYS)
from subprocess import Popen
from hoplite.client.status_updater import MockStatusUpdater
import tempfile
import os
import re


def p4_where_command(client_workspace, p4_path, p4_port, user):
    stdout_file = tempfile.mktemp()
    stdout_file_handle = open(stdout_file, 'w')
    perforce_where_command = ['p4', '-u', user, '-c', client_workspace, '-p', p4_port, 'where', p4_path]
    p4_where = Popen(perforce_where_command, stdout=stdout_file_handle)
    p4_where.communicate()
    p4_where.poll()
    while p4_where.returncode is None:
        p4_where.poll()
    stdout_file_handle.close()
    with open(stdout_file, 'r') as stdout_file_handle:
        p4_where_info = stdout_file_handle.read()
    # TODO: This regex covers windows and unix paths, but is only tested with
    # Unix paths. It needs a Windows test.
    m = re.search(r'((\w:.*$)|\s(/[^/].+))', p4_where_info)
    local_path = m.group().strip()
    return local_path


def p4_sync_command(client_workspace, p4_path, p4_port, user, force=True):
    force_flag = "-f" if force else ""
    perforce_command = ['p4', '-u', user, '-c', client_workspace, '-p', p4_port, 'sync', force_flag, p4_path]
    stdout_file = tempfile.mktemp()
    stdout_file_handle = open(stdout_file, 'w')
    stderr_file = tempfile.mktemp()
    stderr_file_handle = open(stderr_file, 'w')
    p4_sync = Popen(
        perforce_command, stdout=stdout_file_handle, stderr=stderr_file_handle)
    p4_sync.communicate()
    p4_sync.poll()
    while p4_sync.returncode is None:
        p4_sync.poll()
    stdout_file_handle.close()
    stderr_file_handle.close()
    with open(stdout_file, 'r') as stdout_file_handle:
        stdout = stdout_file_handle.read()
    with open(stderr_file, 'r') as stderr_file_handle:
        stderr = stderr_file_handle.read()

    os.remove(stdout_file)
    os.remove(stderr_file)
    return stderr, stdout


def run(config, status):
    user = config.get(KEYS.USER, "")
    client_workspace = config.get(KEYS.CLIENT_WORKSPACE, "")
    p4_port = config.get(KEYS.P4_PORT, "perforce:1666")
    force = config.get(KEYS.FORCE, True)
    p4_path = config.get(KEYS.PERFORCE_PATH, "")

    stderr, stdout = p4_sync_command(
        client_workspace, p4_path, p4_port, user, force=force)

    if stderr == "":
        local_path = p4_where_command(client_workspace, p4_path, p4_port, user)
        if re.search('\.\.\.', local_path):
            # Perforce puts /... if you synced a directory, we need to remove
            # it to have a valid path
            local_path = os.path.dirname(local_path)
        status.update(
            {"succeeded": True, "stdout": stdout, "local_path": local_path})
    else:
        status.update({"succeeded": False, "stdout": stdout, "stderr": stderr})
